#!/usr/bin/env python2.7

from __future__ import absolute_import
from __future__ import print_function
import sys

import numpy as np
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

import tangos as db
from tangos import all_simulations
from tangos import core, config
from tangos.core import Base, get_or_create_dictionary_item, \
    Creator, Simulation, TimeStep, Halo, HaloProperty, HaloLink
from tangos.core.simulation import SimulationProperty
from tangos.core.tracking import TrackData
from tangos.query import get_simulation, get_halo
from tangos.simulation_output_handlers import get_named_handler_class
from tangos.tools.add_simulation import SimulationAdderUpdater
from six.moves import input


def add_simulation_timesteps(options):
    handler=options.handler
    output_class = get_named_handler_class(handler).best_matching_handler(options.sim)
    output_object = output_class(options.sim)
    output_object.quicker = options.quicker
    adder = SimulationAdderUpdater(output_object)
    adder.min_halo_particles = options.min_particles
    adder.scan_simulation_and_add_all_descendants()




def db_import(options):

    sims = options.sims
    remote_db = options.file

    global internal_session
    engine2 = create_engine('sqlite:///' + remote_db, echo=False)
    ext_session = sessionmaker(bind=engine2)()

    _db_import_export(core.get_default_session(), ext_session, *sims)


def db_export(remote_db, *sims):

    global internal_session
    engine2 = create_engine('sqlite:///' + remote_db, echo=False)

    int_session = core.get_default_session()
    ext_session = sessionmaker(bind=engine2)()

    Base.metadata.create_all(engine2)

    _xcurrent_creator = core.creator.get_creator()

    core.set_default_session(ext_session)
    core.set_creator(ext_session.merge(Creator()))

    _db_import_export(ext_session, int_session, *sims)

    core.set_creator(_xcurrent_creator)
    core.set_default_session(int_session)


def _db_import_export(target_session, from_session, *sims):
    from tangos.util.terminalcontroller import heading
    external_to_internal_halo_id = {}
    translated_halolink_ids = []

    if sims == tuple():
        sims = [x.id for x in all_simulations(from_session)]

    for sim in sims:
        ext_sim = get_simulation(sim, from_session)
        sim = Simulation(ext_sim.basename)
        heading("import " + repr(ext_sim))
        sim = target_session.merge(sim)

        for p_ext in ext_sim.properties:
            dic = get_or_create_dictionary_item(
                target_session, p_ext.name.text)
            p = SimulationProperty(sim, dic, p_ext.data)
            p = target_session.merge(p)

        for tk_ext in ext_sim.trackers:
            tk = TrackData(sim, tk_ext.halo_number)
            tk.particles = tk_ext.particles
            tk.use_iord = tk_ext.use_iord

        for ts_ext in ext_sim.timesteps:
            print(".", end=' ')
            sys.stdout.flush()
            ts = TimeStep(sim, ts_ext.extension, False)
            ts.redshift = ts_ext.redshift
            ts.time_gyr = ts_ext.time_gyr
            ts.available = True
            ts = target_session.merge(ts)
            for h_ext in ts_ext.halos:
                h = Halo(ts, h_ext.halo_number, h_ext.finder_id, h_ext.NDM,
                         h_ext.NStar, h_ext.NGas, h_ext.object_typecode)
                h = target_session.merge(h)
                assert h.id is not None and h.id > 0
                external_to_internal_halo_id[h_ext.id] = h.id
                for p_ext in h_ext.properties:
                    dic = get_or_create_dictionary_item(
                        target_session, p_ext.name.text)
                    dat = p_ext.data
                    if dat is not None:
                        p = HaloProperty(h, dic, dat)
                        p = target_session.merge(p)

        print("Translate halolinks", end=' ')
        for ts_ext in ext_sim.timesteps:
            print(".", end=' ')
            sys.stdout.flush()
            _translate_halolinks(
                target_session, ts_ext.links_from, external_to_internal_halo_id, translated_halolink_ids)
            _translate_halolinks(
                target_session, ts_ext.links_to, external_to_internal_halo_id, translated_halolink_ids)

        print("Done")
        target_session.commit()


def _translate_halolinks(target_session, halolinks, external_to_internal_halo_id, translated):
    for hl_ext in halolinks:
        if hl_ext.id in translated:
            continue

        dic = get_or_create_dictionary_item(
            target_session, hl_ext.relation.text)
        hl_new = HaloLink(None, None, dic)

        try:
            hl_new.halo_from_id = external_to_internal_halo_id[
                hl_ext.halo_from_id]
        except KeyError:
            continue

        try:
            hl_new.halo_to_id = external_to_internal_halo_id[hl_ext.halo_to_id]
        except KeyError:
            continue
        target_session.add(hl_new)
        translated.append(hl_ext.id)


def flag_duplicates_deprecated(opts):

    session = db.core.get_default_session()

    print("unmark all:",session.execute("update haloproperties set deprecated=0").rowcount)
    print("      mark:",session.execute("update haloproperties set deprecated=1 where id in (SELECT min(id) FROM haloproperties GROUP BY halo_id, name_id HAVING COUNT(*)>1 ORDER BY halo_id, name_id)").rowcount)

    session.commit()

def remove_duplicates(opts):
    flag_duplicates_deprecated(None)
    count = 1
    while count>0:
        # Order of name_id, halo_id in group clause below is important for optimisation - halo_id, name_id
        # does *not* match the index.
        count = db.core.get_default_session().execute("delete from haloproperties where haloproperties.id in "
                                                 "(SELECT min(id) FROM haloproperties "
                                                 "    GROUP BY name_id, halo_id HAVING COUNT(halo_id)>1);").rowcount
        if count>0 :
            print("Deleted %d rows"%count)
            print("  checking for further duplicates...")
    print("Done")
    db.core.get_default_session().commit()



def rem_simulation_timesteps(options):
    basename = options.sims
    from tangos.util.terminalcontroller import term

    sim = core.get_default_session().query(Simulation).filter_by(
        basename=basename).first()

    if sim == None:
        print(term.GREEN + "Simulation does not exist", term.NORMAL)
    else:
        print(term.RED + "Delete simulation", sim, term.NORMAL)
        core.get_default_session().delete(sim)


def add_tracker(halo, size=None):

    import pynbody

    try:
        halo = get_halo(halo)
    except:

        sim = get_simulation(halo)
        print("Adding tracker for isolated run", sim)
        halo = None

    if halo is not None:
        # check we can open the tracker file
        hfile = halo.load()
        hfile.physical_units()
        use_iord = True
        try:
            hfile.dm['iord']
        except:
            use_iord = False

        # get the centre

        cen = halo.get('SSC', None)
        if cen is None:
            cen = pynbody.analysis.halo.shrink_sphere_center(hfile.dm)

        hfile.ancestor.dm['pos'] -= cen

        if size is None:
            size = '500 pc'

        size = pynbody.units.Unit(size)
        try:
            size.in_units("kpc")
            X = hfile.ancestor.dm[pynbody.filt.Sphere(size)]
        except:
            size.in_units("kpc km s^-1")
            X = hfile.ancestor.dm[pynbody.filt.LowPass("j2", size ** 2)]
            size = str(size.in_units("kpc km s^-1")) + "_kks"

        if len(X) < 2:
            print("Alert! Track data is too short")
            import pdb
            pdb.set_trace()
        # setup the tracker
        tx = TrackData(halo.timestep.simulation)
        print("Tracker halo ID is", tx.halo_number)
        print("Length of track data is", len(X))
        tx.select(X, use_iord)
        ts_trigger = halo.timestep
    else:
        f = sim.timesteps[0].load()
        tx = TrackData(sim)
        if tx.halo_number != 1:
            print("Already have a tracker for this simulation -> abort")
            return
        print("Tracker halo ID is", tx.halo_number)
        tx.particles = np.array(
            np.arange(0, len(f.dm) - 1, 1), dtype=int)
        tx.use_iord = False
        ts_trigger = None

    core.get_default_session().add(tx)
    tx.create_objects(first_timestep=ts_trigger)
    if halo is not None:
        targ = halo.timestep.halos.filter_by(
            object_typecode=1, halo_number=tx.halo_number).first()
        size = str(size).replace(" ", "")
        halo["dm_central_" + size] = targ
        print("Copying SSC...", end=' ')
        sys.stdout.flush()
        while halo is not None:
            try:
                targ['SSC'] = halo['SSC']
            except KeyError:
                pass
            halo = halo.next
            targ = targ.next
        print("done")
    core.get_default_session().commit()


def grep_run(st):
    run = core.get_default_session().query(Creator).filter(
        Creator.command_line.like("%" + st + "%")).all()
    for r in run:
        print(r.id, end=' ')


def list_recent_runs(opts):
    n = opts.num
    run = core.get_default_session().query(Creator).order_by(
        Creator.id.desc()).limit(n).all()
    for r in run:
        r.print_info()


def rem_run(id, confirm=True):
    run = core.get_default_session().query(Creator).filter_by(id=id).first()
    print("You want to delete everything created by the following run:")
    run.print_info()

    if confirm:
        print(""">>> type "yes" to continue""")

    if (not confirm) or input(":").lower() == "yes":
        #for y in run.halolinks:
        #    core.get_default_session().delete(y)
        run.halolinks.delete()
        run.halos.delete()
        run.properties.delete()
        run.timesteps.delete()
        for s in run.simulations:
            print(s)
            core.get_default_session().delete(s)
        core.get_default_session().commit()
        core.get_default_session().delete(run)
        core.get_default_session().commit()
        print("OK")
    else:
        print("aborted")

def rollback(options):
    for run_id in options.ids:
        rem_run(run_id, not options.force)

def dump_id(options):
    import pynbody

    h = db.get_halo(options.halo).load()

    if options.size:
        pynbody.analysis.halo.center(h,vel=False)
        print("Size=",options.size)
        h = h.ancestor[pynbody.filt.Sphere(str(options.size)+" kpc")]

    if options.family!="":
        h = getattr(h,options.family)

    np.savetxt(options.filename,h['iord'],"%d")


def list_available_properties(options):
    from .. import properties
    all_properties = sorted(properties.all_properties())

    def format_class_name(cl):
        return "%s.%s"%(cl.__module__, cl.__name__)

    def format_handler_name(cl):
        if cl.requires_particle_data:
            return cl.works_with_handler.__name__.center(15)
        else:
            return "live".center(15)

    longest_class_name = max([len(format_class_name(cl)) for cl in properties.all_property_classes()])
    print("%s | %s | %s" % ("name".rjust(30), "handler".center(15), "property class"))
    print("-"*30+"-+-"+"-"*15+"-+-"+"-"*longest_class_name)
    for p in all_properties:
        classes = properties.all_providing_classes(p)
        print("%s | %.15s | %s"%(p.rjust(30), format_handler_name(classes[0]), format_class_name(classes[0])))
        for additional_class in classes[1:]:
            print(" "*30+" | %.15s | %s"%(format_handler_name(additional_class),
                                          format_class_name(additional_class)))

def main():

    #db.core.get_default_session() = tangos.blocking_session.BlockingSession(bind = db.core.engine)
    import argparse

    parser = argparse.ArgumentParser()
    core.supplement_argparser(parser)
    parser.add_argument("--verbose", action="store_true",
                        help="Print extra information")


    subparse = parser.add_subparsers()

    subparse_add = subparse.add_parser("add",
                                       help="Add new simulations to the database, or update existing simulations")
    subparse_add.add_argument("sim",action="store",
                              help="The path to the simulation folders relative to the database folder")
    subparse_add.add_argument("--handler", action="store",
                              help="The handler to use from the simulation_outputs subpackage",
                              default=config.default_fileset_handler_class)
    subparse_add.add_argument("--min-particles", action="store", type=int, default=config.min_halo_particles,
                              help="The minimum number of particles a halo must have before it is imported (default %d)"%config.min_halo_particles)
    subparse_add.add_argument("--quicker", action="store_true",
                              help="Cut corners/make guesses to import quickly and with minimum memory usage. Only use if you understand the consequences!")

    subparse_add.set_defaults(func=add_simulation_timesteps)

    subparse_recentruns = subparse.add_parser("recent-runs",
                                              help="List information about the most recent database updates")
    subparse_recentruns.set_defaults(func=list_recent_runs)
    subparse_recentruns.add_argument("num",type=int,
                                     help="The number of runs to display, starting with the most recent")

    subparse_remruns = subparse.add_parser("rm", help="Remove a simulation from the database")
    subparse_remruns.add_argument("sims",help="The path to the simulation folder relative to the database folder")
    subparse_remruns.set_defaults(func=rem_simulation_timesteps)

    subparse_deprecate = subparse.add_parser("flag-duplicates",
                                             help="Flag old copies of properties (if they are present)")
    subparse_deprecate.set_defaults(func=flag_duplicates_deprecated)

    subparse_deprecate = subparse.add_parser("remove-duplicates",
                                             help="Remove old copies of properties (if they are present)")
    subparse_deprecate.set_defaults(func=remove_duplicates)

    subparse_import = subparse.add_parser("import",
                                          help="Import one or more simulations from another sqlite file")
    subparse_import.add_argument("file",type=str,help="The filename of the sqlite file from which to import")
    subparse_import.add_argument("sims",nargs="*",type=str,help="The name of the simulations to import (or import everything if none specified)")
    subparse_import.set_defaults(func=db_import)

    subparse_rollback = subparse.add_parser("rollback", help="Remove database updates (by ID - see recent-runs)")
    subparse_rollback.add_argument("ids",nargs="*",type=int,help="IDs of the database updates to remove")
    subparse_rollback.add_argument("--force","-f",action="store_true",help="Do not prompt for confirmation")
    subparse_rollback.set_defaults(func=rollback)

    subparse_dump_id = subparse.add_parser("dump-iord", help="Dump the iords corresponding to a specified halo")
    subparse_dump_id.add_argument("halo",type=str,help="The identity of the halo to dump")
    subparse_dump_id.add_argument("filename",type=str,help="A filename for the output text file")
    subparse_dump_id.add_argument("size",type=str,nargs="?",help="Size, in kpc, of sphere to extract (or omit to get just the halo particles)")
    subparse_dump_id.add_argument("family",type=str,help="The family of particles to extract",default="")

    subparse_dump_id.set_defaults(func=dump_id)

    subparse_list_available_properties = subparse.add_parser("list-possible-properties", help = "List all the object properties that can be calculated by the currently available modules")
    subparse_list_available_properties.set_defaults(func=list_available_properties)


    args = parser.parse_args()
    core.process_options(args)
    core.init_db()
    args.func(args)



#!/usr/bin/env python
from __future__ import absolute_import
import matplotlib
matplotlib.use('agg')

import Simpy.BlackHoles.orbit, Simpy.BlackHoles.mergers
import sys

def main():
    Simpy.BlackHoles.orbit.sepOrbitbyStep(sys.argv[1], minstep = 1, maxstep = 4096, NCHILADA=False)
    Simpy.BlackHoles.orbit.truncOrbitFile(sys.argv[1], minstep = 1, maxstep = 4096)
    Simpy.BlackHoles.mergers.reducedata(sys.argv[1],outname='changa.out*',mergename='BHmerge.txt',NCHILADA=False)



from __future__ import absolute_import
import contextlib
import six

class MessageMetaClass(type):
    _message_classes = {}

    def __new__(meta, name, bases, dct):
        return super(MessageMetaClass, meta).__new__(meta, name, bases, dct)

    def __init__(cls, name, bases, dct):
        super(MessageMetaClass, cls).__init__(name, bases, dct)
        MessageMetaClass.register_class(cls)

    @staticmethod
    def class_to_hash(cls):
        return hash(cls.__name__) & 0xfffffff

    @staticmethod
    def hash_to_class(hash):
        if hash not in MessageMetaClass._message_classes:
            raise RuntimeError("Unknown message receieved")

        return MessageMetaClass._message_classes[hash]

    @staticmethod
    def class_is_known(cls):
        return MessageMetaClass.class_to_hash(cls) in MessageMetaClass._message_classes

    @staticmethod
    def register_class(cls):
        if MessageMetaClass.class_is_known(cls):
            raise AttributeError("Attempting to register duplicate message class")
        MessageMetaClass._message_classes[MessageMetaClass.class_to_hash(cls)] = cls
        cls._tag = MessageMetaClass.class_to_hash(cls)


class Message(six.with_metaclass(MessageMetaClass, object)):
    _handler = None

    def __init__(self, contents=None):
        self.contents = contents
        self.source = None

    @classmethod
    def deserialize(cls, source, message):
        obj = cls(message)
        obj.source = source
        return obj

    def serialize(self):
        return self.contents

    @staticmethod
    def interpret_and_deserialize(tag, source, message):
        obj = MessageMetaClass.hash_to_class(tag).deserialize(source, message)
        return obj

    @staticmethod
    def process_incoming_message(tag, source, message):
        obj = Message.interpret_and_deserialize(tag, source, message)
        obj.process()

    def send(self, destination):
        from . import backend
        backend.send(self.serialize(), destination=destination, tag=self._tag)

    @classmethod
    def receive(cls, source=None):
        from . import backend

        msg, source, tag = backend.receive_any(source=None)
        obj = Message.interpret_and_deserialize(tag, source, msg)

        if not isinstance(obj, cls):
            if hasattr(obj, "_is_exception"):
                raise obj.contents
            else:
                raise RuntimeError("Unexpected message of type %r received"%type(obj))
        return obj

    def process(self):
        raise NotImplementedError("No process implemented for this message")


class ExceptionMessage(Message):
    _is_exception = True

    def process(self):
        raise self.contents


from __future__ import absolute_import

from mpi4py import MPI
import warnings

import numpy as np

comm = MPI.COMM_WORLD

def send(data, destination, tag=0):
    comm.send(data, dest=destination, tag = tag)

def receive_any(source=None):
    status = MPI.Status()
    if source is None:
        source = MPI.ANY_SOURCE
    data = comm.recv(source=source, tag=MPI.ANY_TAG, status=status)
    return data, status.source, status.tag

def receive(source=None, tag=0):
    if source is None:
        source = MPI.ANY_SOURCE
    return comm.recv(source=source,tag=tag)


def send_numpy_array(data, destination):
    comm.send((data.shape, data.dtype), dest=destination, tag=1)
    comm.Send(data, dest=destination, tag=2)

def receive_numpy_array(source):
    shape,dtype = comm.recv(source=source, tag=1)
    ar = np.empty(shape,dtype=dtype)
    comm.Recv(ar,source=source,tag=2)
    return ar

def rank():
    return comm.Get_rank()

def size():
    return comm.Get_size()

def barrier():
    comm.Barrier()

def finalize():
    MPI.Finalize()


def launch(function, num_procs, args):
    if size()==1:
        raise RuntimeError("MPI run needs minimum of 2 processors (one for manager, one for worker)")
    if num_procs is not None:
        if rank()==0:
            warnings.warn("Number of processors requested (%d) will be ignored as this is an MPI run that has already selected %d processors"%(num_procs,size()))
    function(*args)


"""tangos - the amazing numerical galaxy organization system

This package provides ways to create and interpret databases of numerical galaxy formation simulations for scientific
analysis.

For information on getting started, see README.md.

"""

from __future__ import absolute_import
import sqlalchemy
import sqlalchemy.orm.session
from sqlalchemy import Index, Column, Integer, String, Float, ForeignKey, DateTime, Boolean, LargeBinary, create_engine, orm
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, backref, sessionmaker, clear_mappers, deferred
from sqlalchemy import and_, or_
from sqlalchemy.orm.session import Session




from . import core, log
from .core import *
from .query import *
from . import properties

core.init_db()

__version__ = "1.0.0"


from .multi_hop import MultiHopStrategy
from ..config import num_multihops_max_default as NHOPS_MAX_DEFAULT


class MultiHopAllProgenitorsStrategy(MultiHopStrategy):
    """Finds all progenitors for a halo at every step"""
    def __init__(self, halo_from, nhops_max=NHOPS_MAX_DEFAULT, include_startpoint=False, target='auto', combine_routes=True):
        self.sim_id = halo_from.timestep.simulation_id
        if target=='auto':
            target = halo_from.timestep.simulation
        super(MultiHopAllProgenitorsStrategy, self).__init__(halo_from, nhops_max,
                                                               directed='backwards',
                                                               include_startpoint=include_startpoint,
                                                               target=target,
                                                               order_by=['time_desc', 'halo_number_asc'],
                                                               combine_routes=combine_routes,
                                                             min_onehop_reverse_weight=0.1)

    def _supplement_halolink_query_with_filter(self, query, table):
        query = super(MultiHopAllProgenitorsStrategy, self)._supplement_halolink_query_with_filter(query, table)
        if self._target is None:
            return query
        else:
            return query.filter(self.timestep_new.simulation_id == self.sim_id)


class MultiHopMajorProgenitorsStrategy(MultiHopAllProgenitorsStrategy):
    """Finds the major progenitor for a halo at every step"""

    def _supplement_halolink_query_with_filter(self, query, table):
        query = super(MultiHopMajorProgenitorsStrategy, self)._supplement_halolink_query_with_filter(query, table)
        return query.order_by(self.timestep_new.time_gyr.desc(), table.c.weight.desc(), self.halo_new.halo_number). \
            limit(1)

class MultiHopMostRecentMergerStrategy(MultiHopAllProgenitorsStrategy):
    """Finds the halos involved in the most recent merger into the major progenitor branch of the halo"""

    def _hopping_finished(self, filtered_count):
        self._last_filtered_count = filtered_count
        return filtered_count != 1

    def _make_hops(self):
        super(MultiHopMostRecentMergerStrategy, self)._make_hops()
        if self._last_filtered_count>1:
            self.query = self.query.filter(self._table.c.nhops==self._nhops_taken+1)
        else:
            # no merger was found
            self.query = self.query.filter(0==1)



class MultiHopMajorDescendantsStrategy(MultiHopStrategy):
    """Suggests the major descendant for a halo at every step"""

    def __init__(self, halo_from, nhops_max=NHOPS_MAX_DEFAULT, include_startpoint=False):
        self.sim_id = halo_from.timestep.simulation_id
        super(MultiHopMajorDescendantsStrategy, self).__init__(halo_from, nhops_max,
                                                               directed='forwards',
                                                               include_startpoint=include_startpoint,
                                                               target=halo_from.timestep.simulation)

    def _supplement_halolink_query_with_filter(self, query, table):
        query = super(MultiHopMajorDescendantsStrategy, self)._supplement_halolink_query_with_filter(query, table)
        return query.filter(self.timestep_new.simulation_id == self.sim_id). \
            order_by(self.timestep_new.time_gyr, table.c.weight.desc(), self.halo_new.halo_number). \
            limit(1)





from __future__ import absolute_import
import sqlalchemy, sqlalchemy.engine, sqlalchemy.event
from . import parallel_tasks as pt
from . import core
global rlock

__author__ = 'app'

def event_begin(conn):
    rlock.acquire()

def event_commit_or_rollback(conn):
    rlock.release()


def make_engine_blocking(engine=None):
    global rlock
    if engine is None:
        engine = core.get_default_engine()
    assert isinstance(engine, sqlalchemy.engine.Engine)
    rlock = pt.RLock("db_write_lock")

    sqlalchemy.event.listen(engine, 'begin', event_begin )
    sqlalchemy.event.listen(engine, 'commit', event_commit_or_rollback)
    sqlalchemy.event.listen(engine, 'rollback', event_commit_or_rollback)


from __future__ import absolute_import
from tangos.core import extraction_patterns
from . import BuiltinFunction
from .. import StoredProperty, FixedInput


@BuiltinFunction.register
def raw(halos, values):
    return values
raw.set_input_options(0, assert_class=StoredProperty)

@raw.set_initialisation
def raw_initialisation(input):
    input.set_extraction_pattern(extraction_patterns.halo_property_raw_value_getter)


@BuiltinFunction.register
def reassemble(halos, values, *options):
    return values
reassemble.set_input_options(0, assert_class=StoredProperty)

@reassemble.set_initialisation
def reassemble_initialisation(input, *options):
    options_values = []
    for option in options:
        if isinstance(option, FixedInput):
            options_values.append(option.proxy_value())
        else:
            raise TypeError("Options to 'reassemble' must be fixed numbers or strings")

    input.set_extraction_pattern(
        extraction_patterns.HaloPropertyValueWithReassemblyOptionsGetter(*options_values))



"""Logic to help flexibly get/set data in ORM objects (specifically HaloProperty and SimulationProperty) where data
of different types has to be stored in different attributes.
"""

from __future__ import absolute_import
import numpy as np
import zlib
import time
import datetime
import six
import sys
import functools
from six.moves import cPickle as pickle

pickle_loads = pickle.loads
if int(sys.version[0])==3:
    pickle_loads = functools.partial(pickle.loads, encoding='latin1')


_THRESHOLD_FOR_COMPRESSION = 1000

def get_data_of_unknown_type(obj):
    """Starting from the ORM object, extract data which may be stored in a variety of attributes depending on its type"""
    mapper = DataAttributeMapper(db_object=obj)
    return mapper.get(obj)

def set_data_of_unknown_type(obj, data):
    """Store the given data in the ORM object, using an appropriate attribute for the given data type"""
    mapper = DataAttributeMapper(data=data)
    mapper.set(obj,data)


class DataAttributeMapper(object):
    _order = 0
    # this can be used to force a subclass to be 'found' last
    # see __all_subclasses below, and use in NullAttributeMapper
    #
    # Except for a particular branch of py 3.5 it seems that __subclasses__ always
    # returns in order that classes were defined anyway, so it is not clear why this
    # is necessary.

    def __new__(cls, db_object=None, data=None):
        if db_object is None and data is None:
            subclass=NullAttributeMapper
        if data is None:
            subclass=cls._subclass_from_db_object(db_object)
        else:
            subclass=cls._subclass_from_data(data)

        return object.__new__(subclass)

    @classmethod
    def __all_subclasses(cls):
        subclasses = cls.__subclasses__()
        subclasses.sort(key = lambda x: x._order)
        for X in subclasses:
            yield X
            for Y in X.__all_subclasses():
                yield Y

    @classmethod
    def __all_nonabstract_subclasses(cls):
        X= [sub for sub in cls.__all_subclasses() if hasattr(sub,"_attribute_name")]
        return X

    @classmethod
    def _subclass_from_db_object(cls, db_object):
        for subclass in cls.__all_nonabstract_subclasses():
            if subclass._handles_db_object(db_object):
                return subclass

        # NullAttributeMapper should prevent the following line from being reached
        assert False, "Internal error in data_attribute_mapper: no subclass claimed responsibility for object"

    @classmethod
    def _subclass_from_data(cls, data):
        for subclass in cls.__all_nonabstract_subclasses():
            if subclass._handles_data(data):
                return subclass
        raise TypeError("Don't know how to store data of type %r"%type(data))

    @classmethod
    def _handles_db_object(cls, db_object):
        if cls._attribute_name is None:
            return True
        return getattr(db_object, cls._attribute_name, None) is not None

    @classmethod
    def _handles_data(cls, data):
        return any([isinstance(data, t) for t in cls._handled_types])

    def pack(self, data):
        return data

    def unpack(self,data):
        return data

    def _clear_other_attributes(self, db_object):
        for cls in DataAttributeMapper.__all_nonabstract_subclasses():
            if not isinstance(self, cls) and cls._attribute_name is not None:
                setattr(db_object, cls._attribute_name, None)

    def set(self, db_object, data):
        if hasattr(db_object,self._attribute_name):
            setattr(db_object, self._attribute_name, self.pack(data))
        else:
            raise TypeError("%r object does not have a slot for %r"%(type(db_object),self._attribute_name))

        self._clear_other_attributes(db_object)

    def get(self, db_object):
        return self.unpack(getattr(db_object, self._attribute_name))

class TimeAttributeMapper(DataAttributeMapper):
    _attribute_name = "data_time"
    _handled_types = [datetime.datetime, time.struct_time]

    def pack(self, data):
        if isinstance(data, time.struct_time):
            data = datetime.datetime(*data[:6])
        return data

class StringAttributeMapper(DataAttributeMapper):
    _attribute_name = "data_string"
    _handled_types = [str, six.text_type]

class ArrayDowncastingAttributeMapper(DataAttributeMapper):
    @classmethod
    def _handles_data(cls, data):
        if any([isinstance(data, t) for t in cls._handled_types]):
            return True
        if isinstance(data, list) and len(data)==1 and cls._handles_data(data[0]):
            return True
        if isinstance(data, np.ndarray) and len(data.shape)==0 and \
            any([data.dtype == t for t in cls._handled_types]):
            return True
        return False

    def pack(self, data):
        if isinstance(data,list):
            return data[0]
        else:
            return self._handled_types[0](data)

class FloatAttributeMapper(ArrayDowncastingAttributeMapper):
    _attribute_name = "data_float"
    _handled_types = [float, np.float32, np.float64, np.float128]


class IntAttributeMapper(ArrayDowncastingAttributeMapper):
    _attribute_name = "data_int"
    _handled_types = [int, np.int32, np.int64]


class ArrayAttributeMapper(DataAttributeMapper):
    _attribute_name = "data_array"
    _handled_types = [list, np.ndarray]
    _order = 1 # must be used only when downcasting mappers have failed

    def _unpack_compressed(self, packed):
        return pickle_loads(zlib.decompress(packed[2:]))

    def _unpack_uncompressed(self, packed):
        return pickle_loads(packed[2:])

    def _unpack_old_format(self, packed):
        return np.frombuffer(packed)

    def unpack(self, packed):
        if len(packed)==0:
            return None
        elif packed.startswith(b"ZX"):
            return self._unpack_compressed(packed)
        elif packed.startswith(b"PX"):
            return self._unpack_uncompressed(packed)
        else:
            return self._unpack_old_format(packed)

    def pack(self, data):
        dumped_st = pickle.dumps(data)
        if len(dumped_st) > _THRESHOLD_FOR_COMPRESSION:
            dumped_st = b"ZX" + zlib.compress(dumped_st)
        else:
            dumped_st = b"PX" + dumped_st
        return dumped_st


# Following must be defined last to act as a fall-through:
class NullAttributeMapper(DataAttributeMapper):
    _handled_types = [type(None)]
    _attribute_name = None
    _order = 2 # see discussion in the DataAttributeMapper

    def set(self, db_object, data):
        assert data is None
        self._clear_other_attributes(db_object)

    def get(self, db_object):
        return None

__all__ = ['get_data_of_unknown_type', 'set_data_of_unknown_type']



from __future__ import absolute_import
from .message import Message, ExceptionMessage
from . import log, parallel_backend_loaded, remote_import
from ..util.check_deleted import check_deleted
import sys
import pynbody
import gc
import six.moves.cPickle as pickle
import numpy as np
from six.moves import zip


currently_loaded_snapshot = None

class ConfirmLoadPynbodySnapshot(Message):
    pass


class PynbodySnapshotQueue(object):
    def __init__(self):
        self.load_file_queue = []
        self.load_requester_queue = []
        self.current_snapshot_filename = None
        self.current_snapshot = None
        self.current_subsnap_cache = {}
        self.current_halocat = None
        self.in_use_by = []

    def halos(self):
        if self.current_halocat is None:
            self.current_halocat = self.current_snapshot.halos()
        return self.current_halocat

    def add(self, filename, requester):
        log.logger.debug("Pynbody server: client %d requests access to %r", requester, filename)
        if filename==self.current_snapshot_filename:
            self._notify_available(requester)
            self.in_use_by.append(requester)
        elif filename in self.load_file_queue:
            self.load_requester_queue[self.load_file_queue.index(filename)].append(requester)
        else:
            self.load_file_queue.append(filename)
            self.load_requester_queue.append([requester])
        self._load_next_if_free()

    def free(self, requester):
        self.in_use_by.remove(requester)
        log.logger.debug("Pynbody server: client %d is now finished with %r", requester, self.current_snapshot_filename)
        self._free_if_unused()
        self._load_next_if_free()

    def get_subsnap(self, filter_, fam):
        if (filter_,fam) in self.current_subsnap_cache:
            return self.current_subsnap_cache[(filter_, fam)]
        else:
            subsnap = self.get_subsnap_uncached(filter_, fam)
            self.current_subsnap_cache[(filter_, fam)] = subsnap
            return subsnap

    def get_subsnap_uncached(self, filter_, fam):

        snap = self.current_snapshot

        if isinstance(filter_, pynbody.filt.Filter):
            snap = snap[filter_]
        else:
            snap = self.halos()[filter_]

        if fam is not None:
            snap = snap[fam]

        return snap



    def _free_if_unused(self):
        if len(self.in_use_by)==0:
            log.logger.debug("Pynbody server: all clients are finished with the current snapshot; freeing.")
            with check_deleted(self.current_snapshot):
                self.current_snapshot = None
                self.current_snapshot_filename = None
                self.current_halocat = None
                self.current_subsnap_cache = {}

    def _notify_available(self, node):
        log.logger.debug("Pynbody server: notify %d that snapshot is now available", node)
        ConfirmLoadPynbodySnapshot(type(self.current_snapshot)).send(node)

    def _load_next_if_free(self):
        if len(self.load_file_queue)==0:
            return

        if self.current_snapshot_filename is None:
            # TODO: Error handling
            self.current_snapshot_filename = self.load_file_queue.pop(0)
            self.current_snapshot = pynbody.load(self.current_snapshot_filename)
            self.current_snapshot.physical_units()
            log.logger.info("Pynbody server: loaded %r", self.current_snapshot_filename)

            notify = self.load_requester_queue.pop(0)
            self.in_use_by = notify
            for n in notify:
                self._notify_available(n)
        else:
            log.logger.info("The currently loaded snapshot is still required and so other clients will have to wait")
            log.logger.info("(Currently %d snapshots are in the queue to be loaded later)", len(self.load_file_queue))



_server_queue = PynbodySnapshotQueue()

class RequestLoadPynbodySnapshot(Message):
    def process(self):
        _server_queue.add(self.contents, self.source)

class ReleasePynbodySnapshot(Message):
    def process(self):
        _server_queue.free(self.source)

class ReturnPynbodyArray(Message):
    @classmethod
    def deserialize(cls, source, message):
        from . import backend
        contents = backend.receive_numpy_array(source=source)

        if message!="":
            contents = contents.view(pynbody.array.SimArray)
            contents.units = pickle.loads(message)

        obj = ReturnPynbodyArray(contents)
        obj.source = source

        return obj

    def serialize(self):
        assert isinstance(self.contents, np.ndarray)
        if hasattr(self.contents, 'units'):
            serialized_info = pickle.dumps(self.contents.units)
        else:
            serialized_info = ""

        return serialized_info

    def send(self, destination):
        # send envelope
        super(ReturnPynbodyArray, self).send(destination)

        # send contents
        from . import backend
        backend.send_numpy_array(self.contents.view(np.ndarray), destination)

class RequestPynbodyArray(Message):
    def __init__(self, filter_, array, fam=None):
        self.filter_ = filter_
        self.array = array
        self.fam = fam

    @classmethod
    def deserialize(cls, source, message):
        obj = RequestPynbodyArray(*message)
        obj.source = source
        return obj

    def serialize(self):
        return (self.filter_, self.array, self.fam)

    def process(self):
        with _server_queue.current_snapshot.immediate_mode, _server_queue.current_snapshot.lazy_derive_off:
            try:
                subsnap = _server_queue.get_subsnap(self.filter_, self.fam)
                if subsnap._array_name_implies_ND_slice(self.array):
                    raise KeyError("Not transferring a single slice %r of a ND array"%self.array)
                if self.array=='remote-index-list':
                    subarray = subsnap.get_index_list(subsnap.ancestor)
                else:
                    subarray = subsnap[self.array]
                    assert isinstance(subarray, pynbody.array.SimArray)
                array_result = ReturnPynbodyArray(subarray)
            except Exception as e:
                array_result = ExceptionMessage(e)
            array_result.send(self.source)
            del array_result
            gc.collect()


class ReturnPynbodySubsnapInfo(Message):
    def __init__(self, families, sizes, properties, loadable_keys, fam_loadable_keys):
        super(ReturnPynbodySubsnapInfo, self).__init__()
        self.families = families
        self.sizes = sizes
        self.properties = properties
        self.loadable_keys = loadable_keys
        self.fam_loadable_keys = fam_loadable_keys

    def serialize(self):
        return self.families, self.sizes, self.properties, self.loadable_keys, self.fam_loadable_keys

    @classmethod
    def deserialize(cls, source, message):
        obj = cls(*message)
        obj.source = source
        return obj



class RequestPynbodySubsnapInfo(Message):
    def __init__(self, filename, filter_):
        super(RequestPynbodySubsnapInfo, self).__init__()
        self.filename = filename
        self.filter_ = filter_

    @classmethod
    def deserialize(cls, source, message):
        obj = cls(*message)
        obj.source = source
        return obj

    def serialize(self):
        return (self.filename, self.filter_)

    def process(self):
        assert(_server_queue.current_snapshot_filename==self.filename)
        if isinstance(self.filter_, pynbody.filt.Filter):
            obj = _server_queue.current_snapshot[self.filter_]
        else:
            obj = _server_queue.halos()[self.filter_]
        families = obj.families()
        fam_lengths = [len(obj[fam]) for fam in families]
        fam_lkeys = [obj.loadable_keys(fam) for fam in families]
        lkeys = obj.loadable_keys()
        ReturnPynbodySubsnapInfo(families, fam_lengths, obj.properties, lkeys, fam_lkeys).send(self.source)


class RemoteSubSnap(pynbody.snapshot.SimSnap):
    def __init__(self, connection, filter_):
        super(RemoteSubSnap, self).__init__()

        self.connection = connection
        self._filename = connection.identity
        self._server_id = connection._server_id

        RequestPynbodySubsnapInfo(connection.filename, filter_).send(self._server_id)
        info = ReturnPynbodySubsnapInfo.receive(self._server_id)

        index = 0
        for fam, size in zip(info.families, info.sizes):
            self._family_slice[fam] = slice(index, index+size)
            index+=size
        self._num_particles = index

        self.properties.update(info.properties)
        self._loadable_keys = info.loadable_keys
        self._fam_loadable_keys = {fam: lk for fam, lk in zip(info.families, info.fam_loadable_keys)}
        self.filter_ = filter_

    def _find_deriving_function(self, name):
        cl = self.connection.underlying_pynbody_class
        if cl in self._derived_quantity_registry \
                and name in self._derived_quantity_registry[cl]:
            return self._derived_quantity_registry[cl][name]
        else:
            return super(RemoteSubSnap, self)._find_deriving_function(name)


    def _load_array(self, array_name, fam=None):
        RequestPynbodyArray(self.filter_,array_name,fam).send(self._server_id)
        try:
            data = ReturnPynbodyArray.receive(self._server_id).contents
        except KeyError:
            raise IOError("No such array %r available from the remote"%array_name)
        if fam is None:
            self[array_name] = data
        else:
            self[fam][array_name] = data


_connection_active = False

class RemoteSnapshotConnection(object):
    def __init__(self, fname, server_id=0):
        global _connection_active

        if _connection_active:
            raise RuntimeError("Each client can only have one remote snapshot connection at any time")

        _connection_active = True

        super(RemoteSnapshotConnection, self).__init__()

        self._server_id = 0
        self.filename = fname
        self.identity = "%d: %s"%(self._server_id,fname)

        # ensure server knows what our messages are about
        remote_import.ImportRequestMessage(__name__).send(self._server_id)

        log.logger.debug("Pynbody client: attempt to connect to remote snapshot %r", fname)
        RequestLoadPynbodySnapshot(fname).send(self._server_id)
        self.underlying_pynbody_class = ConfirmLoadPynbodySnapshot.receive(self._server_id).contents
        self.connected = True

        log.logger.info("Pynbody client: connected to remote snapshot %r", fname)

    def get_view(self, filter_):
        return RemoteSubSnap(self, filter_)

    def disconnect(self):
        global _connection_active

        if not self.connected:
            return

        ReleasePynbodySnapshot(self.filename).send(self._server_id)
        _connection_active = False
        self.connected = False

    def __del__(self):
        self.disconnect()









from __future__ import absolute_import
from __future__ import print_function
import tangos
from . import core
import sqlalchemy, sqlalchemy.event
import contextlib
import gc
import traceback
import os
import inspect
import six
from six.moves import range
from six.moves import zip

class TestSimulationGenerator(object):
    def __init__(self, sim_name="sim", session=None):
        if not session:
            session = core.get_default_session()

        self.session = session
        self.sim = tangos.core.simulation.Simulation(sim_name)
        self._ptcls_in_common_dict = core.dictionary.get_or_create_dictionary_item(self.session, "ptcls_in_common")
        self._BH_dict = core.dictionary.get_or_create_dictionary_item(self.session, "BH")
        self._host_dict = core.dictionary.get_or_create_dictionary_item(self.session, "host")
        self.session.commit()

    def add_objects_to_timestep(self, num_halos, object_typecode=0, NDM=None):
        """Add halos to the most recently added timestep"""

        ts = self._most_recently_added_timestep()
        returned_halos = []

        for i in range(1,num_halos+1):
            if NDM is None:
                NDM_halo = 1000-i*100
            else:
                NDM_halo = NDM[i-1]
            halo = tangos.core.halo.Halo(ts, i, i, NDM_halo, 0, 0, 0)
            halo.object_typecode = object_typecode
            returned_halos.append(self.session.merge(halo))

        self.session.commit()
        return returned_halos

    def add_properties_to_halos(self, **properties):
        """Add properties to the halos in the most recently added timestep.

        For each kwarg, pass in a function mapping the halo number to the named property

        e.g. add_properties_to_halos(Mvir=lambda i: 100.*(10-i)) adds a Mvir value of 100*(10-halo_number)
         to each halo"""

        self._add_properties_to_objects(0, **properties)

    def _add_properties_to_objects(self, htype, **properties):
        """Add properties to the halos in the most recently added timestep.

        For each kwarg, pass in a function mapping the halo number to the named property

        e.g. add_properties_to_halos(Mvir=lambda i: 100.*(10-i)) adds a Mvir value of 100*(10-halo_number)
         to each halo"""

        for k, v in six.iteritems(properties):
            for halo in self._most_recently_added_timestep().objects.filter_by(object_typecode=htype).all():
                halo[k] = v(halo.halo_number)

    def add_bhs_to_timestep(self, num_bhs):
        halos = self.add_objects_to_timestep(num_bhs, 1)
        return halos

    def add_properties_to_bhs(self, **properties):
        """Add properties to BHs. See add_properties_to_halos for documentation"""

        self._add_properties_to_objects(1, **properties)

    def _most_recently_added_timestep(self):
        return self.sim.timesteps[-1]

    def _two_most_recently_added_timesteps(self):
        return self.sim.timesteps[-2:]

    def add_timestep(self):
        """Add a sequentially-numbered timestep to the specified simulation"""

        timestep_num = len(self.sim.timesteps)+1
        ts = tangos.core.timestep.TimeStep(self.sim, "ts%d"%timestep_num)
        ts.redshift = 9 - timestep_num
        ts.time_gyr = 0.9*timestep_num
        self.session.add(ts)
        self.session.commit()

        return ts

    def link_last_halos_using_mapping(self, mapping, consistent_masses=True, object_typecode=0) :
        """Store halolinks such that ts_source[halo_i].next = ts_dest[mapping[halo_i]]

        :type mapping dict"""

        ts_source, ts_dest = self._two_most_recently_added_timesteps()
        if consistent_masses:
            self._adjust_halo_NDM_for_mapping(mapping, ts_dest, ts_source)
        self._generate_bidirectional_halolinks_for_mapping(mapping, ts_dest, ts_source, object_typecode)
        self.session.commit()

    def link_last_halos(self, object_typecode=0):
        """Generate default halolinks for the most recent two timesteps such that 1->1, 2->2 etc"""

        ts_source, ts_dest = self._two_most_recently_added_timesteps()
        halo_nums_source = set([a.halo_number for a in ts_source.objects.filter_by(object_typecode=object_typecode).all()])
        halo_nums_dest = set([a.halo_number for a in ts_dest.objects.filter_by(object_typecode=object_typecode).all()])
        halo_nums_common = halo_nums_source.intersection(halo_nums_dest)
        mapping = dict([(a,a) for a in halo_nums_common])
        self.link_last_halos_using_mapping(mapping,False, object_typecode)


    def link_last_bhs(self):
        self.link_last_halos(1)

    def link_last_bhs_using_mapping(self, mapping):
        self.link_last_halos_using_mapping(mapping, object_typecode=1)

    def assign_bhs_to_halos(self, mapping):
        ts = self._most_recently_added_timestep()

        for k,v in six.iteritems(mapping):
            source_halo = ts.halos.filter_by(halo_number=v).first()
            target_bh = ts.bhs.filter_by(halo_number=k).first()

            forward_link = core.halo_data.HaloLink(source_halo, target_bh, self._BH_dict,
                                                   1.0)
            self.session.add(forward_link)

            reverse_link = core.halo_data.HaloLink(target_bh, source_halo, self._host_dict,
                                                   1.0)
            self.session.add(reverse_link)

        self.session.commit()

    def link_last_halos_across_using_mapping(self, other, mapping, object_typecode=0):
        ts_source = self._most_recently_added_timestep()
        ts_dest = other._most_recently_added_timestep()
        self._generate_bidirectional_halolinks_for_mapping(mapping, ts_dest, ts_source, object_typecode)
        self.session.commit()

    def add_mass_transfer(self, source_num, target_num, fraction_of_source, object_typecode=0):
        ts_source, ts_dest = self._two_most_recently_added_timesteps()
        source_halo = ts_source.objects.filter_by(halo_number=source_num, object_typecode=object_typecode).first()
        target_halo = ts_dest.objects.filter_by(halo_number=target_num, object_typecode=object_typecode).first()

        transferred_ptcls = int(source_halo.NDM * fraction_of_source)
        fraction_of_dest = float(transferred_ptcls)/target_halo.NDM
        ts_source, ts_dest = self._two_most_recently_added_timesteps()

        forward_link = core.halo_data.HaloLink(source_halo, target_halo, self._ptcls_in_common_dict, fraction_of_source)
        backward_link = core.halo_data.HaloLink(target_halo, source_halo, self._ptcls_in_common_dict, fraction_of_dest)


    def _generate_bidirectional_halolinks_for_mapping(self, mapping, ts_dest, ts_source, object_typecode):

        for source_num, target_num in six.iteritems(mapping):
            source_halo = ts_source.objects.filter_by(halo_number=source_num, object_typecode=object_typecode).first()
            target_halo = ts_dest.objects.filter_by(halo_number=target_num, object_typecode=object_typecode).first()

            forward_link = core.halo_data.HaloLink(source_halo, target_halo, self._ptcls_in_common_dict, 1.0)
            backward_link = core.halo_data.HaloLink(target_halo, source_halo, self._ptcls_in_common_dict, float(source_halo.NDM) / target_halo.NDM)

            self.session.add_all([forward_link, backward_link])


    def _adjust_halo_NDM_for_mapping(self, mapping, ts_dest, ts_source):

        for source_num, target_num in six.iteritems(mapping):
            target_halo = ts_dest.halos.filter_by(halo_number=target_num).first()
            target_halo.NDM = 0
        for source_num, target_num in six.iteritems(mapping):
            source_halo = ts_source.halos.filter_by(halo_number=source_num).first()
            target_halo = ts_dest.halos.filter_by(halo_number=target_num).first()
            print(source_num, ts_source)
            target_halo.NDM += source_halo.NDM



def _as_halos(hlist, session=None):
    if session is None:
        session = core.get_default_session()
    rvals = []
    for h in hlist:
        if h is None:
            rvals.append(None)
        elif isinstance(h, core.halo.Halo):
            rvals.append(h)
        else:
            rvals.append(tangos.get_halo(h, session))
    return rvals

def _halos_to_strings(hlist):
    if len(hlist)==0:
        return "(empty list)"
    else:
        return str([hx.path if hx else "None" for hx in _as_halos(hlist)])

def halolists_equal(hl1, hl2, session=None):
    """Return True if hl1 and hl2 are equivalent lists of halos"""

    hl1 = _as_halos(hl1)
    hl2 = _as_halos(hl2)

    return len(hl1)==len(hl2) and all([h1==h2 for h1, h2 in zip(hl1,hl2)])

def assert_halolists_equal(hl1, hl2, session=None):
    equal = halolists_equal(hl1, hl2, session=None)
    assert equal, "Not equal: %s %s"%(_halos_to_strings(hl1),_halos_to_strings(hl2))

@contextlib.contextmanager
def autorevert():
    old_session = core.get_default_session()
    connection = core.get_default_engine().connect()
    transaction = connection.begin()
    isolated_session = core.Session(bind=connection)
    core.set_default_session(isolated_session)
    yield
    transaction.rollback()
    core.set_default_session(old_session)




@contextlib.contextmanager
def assert_connections_all_closed():
    num_connections = [0,0]
    connection_details = {}
    def on_checkout(dbapi_conn, connection_rec, connection_proxy):
        num_connections[0]+=1
        num_connections[1]+=1
        connection_details[id(connection_rec)] = traceback.extract_stack()

    def on_checkin(dbapi_conn, connection_rec):
        if id(connection_rec) in connection_details:
            num_connections[0]-=1
            del connection_details[id(connection_rec)]

    gc.collect()

    sqlalchemy.event.listen(core.get_default_engine().pool, 'checkout', on_checkout)
    sqlalchemy.event.listen(core.get_default_engine().pool, 'checkin', on_checkin)

    yield


    gc.collect()

    sqlalchemy.event.remove(core.get_default_engine().pool, 'checkout', on_checkout)
    sqlalchemy.event.remove(core.get_default_engine().pool, 'checkin', on_checkin)

    for k,v in six.iteritems(connection_details):
        print("object id",k,"not checked in; was created here:")
        for line in traceback.format_list(v):
            print("  ",line)

    assert num_connections[0]==0, "%d (of %d) connections were not closed"%(num_connections[0], num_connections[1])


def init_blank_db_for_testing(**init_kwargs):
    try:
        os.mkdir("test_dbs")
    except OSError:
        pass

    caller_fname = os.path.basename(inspect.getframeinfo(inspect.currentframe().f_back)[0])[:-3]

    print(caller_fname)
    db_name = "test_dbs/%s.db"%caller_fname
    try:

        os.remove(db_name)
    except OSError:
        pass

    core.init_db("sqlite:///"+db_name,**init_kwargs)




from __future__ import absolute_import
from __future__ import print_function
from pyramid.view import view_config
from . import halo_from_request
from ...relation_finding import tree

class WebMergerTree(tree.MergerTree):
    def __init__(self, halo, request):
        self.request = request
        super(WebMergerTree, self).__init__(halo)

    def _get_basic_halo_node(self, halo, depth):
        output = super(WebMergerTree, self)._get_basic_halo_node(halo, depth)
        output['url'] = self.request.route_url('halo_view',
                                       simid=self.base_halo.timestep.simulation.basename,
                                       timestepid=halo.timestep.extension,
                                       halonumber=halo.basename)
        return output

def _construct_mergertree(halo, request):
    base = halo
    must_include = []
    for i in range(5):
        must_include.append(base.id)
        if base.next is not None:
            base = base.next

    tree = WebMergerTree(base, request)
    tree.x_step = 30
    tree.must_include = must_include
    tree.highlight_halo = halo
    tree.construct()
    return tree._treedata


@view_config(route_name='merger_tree', renderer='json')
def merger_tree(request):
    halo = halo_from_request(request)
    return {'tree': _construct_mergertree(halo, request)}


from __future__ import absolute_import
import weakref

from sqlalchemy import Column, Integer, ForeignKey, orm
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.declarative import declared_attr

from . import extraction_patterns
from . import Base
from . import creator
from .dictionary import get_dict_id, get_or_create_dictionary_item
from .timestep import TimeStep
import six

class Halo(Base):
    __tablename__= "halos"

    id = Column(Integer, primary_key=True)
    halo_number = Column(Integer)
    finder_id = Column(Integer)
    timestep_id = Column(Integer, ForeignKey('timesteps.id'))
    timestep = relationship(TimeStep, backref=backref(
        'objects', order_by=halo_number, cascade='all', lazy='dynamic'), cascade='save-update, merge')
    NDM = Column(Integer)
    NStar = Column(Integer)
    NGas = Column(Integer)
    creator = relationship(creator.Creator, backref=backref(
        'halos', cascade='all', lazy='dynamic'), cascade='save-update')
    creator_id = Column(Integer, ForeignKey('creators.id'))
    object_typecode = Column(Integer, nullable=False,
                         name='halo_type') # name for backwards compatibility

    tag = "halo" # used identifying this default object type, as opposed to other object types,
                 # e.g. a group/BH defined below

    __mapper_args__ = {
        'polymorphic_identity':0,
        'polymorphic_on':object_typecode
    }

    @classmethod
    def _all_subclasses(cls):
        for c in cls.__subclasses__():
            yield c
            for c_sub in c._all_subclasses():
                yield c_sub

    @staticmethod
    def class_from_tag(match_tag):
        match_tag = match_tag.lower()
        if match_tag == Halo.tag.lower():
            return Halo
        for c in Halo._all_subclasses():
            if match_tag == c.tag.lower():
                return c
        raise ValueError("Unknown object type %r"%match_tag)

    @staticmethod
    def object_typecode_from_tag(match_tag):
        if isinstance(match_tag, six.string_types):
            return Halo.class_from_tag(match_tag).__mapper_args__['polymorphic_identity']
        else:
            return match_tag

    @staticmethod
    def object_typetag_from_code(typecode):
        if typecode==0:
            return Halo.tag
        else:
            for c in Halo._all_subclasses():
                if c.__mapper_args__['polymorphic_identity'] == typecode:
                    return c.tag
        raise ValueError("Unknown object typecode %d",typecode)

    def __init__(self, timestep, halo_number, finder_id, NDM, NStar, NGas, object_typecode=0):
        self.timestep = timestep
        self.halo_number = int(halo_number)
        self.finder_id = int(finder_id)
        self.NDM = int(NDM)
        self.NStar = int(NStar)
        self.NGas = int(NGas)
        self.object_typecode = int(object_typecode)
        self.init_on_load()
        self.creator_id = creator.get_creator_id()

    @orm.reconstructor
    def init_on_load(self):
        self._dict_is_complete = False
        self._d = {}

    def __repr__(self):

        return "<%s %r | NDM=%d Nstar=%d Ngas=%d>"%(self.__class__.__name__, self.path, self.NDM, self. NStar, self.NGas)


    @property
    def basename(self):
        return self.tag+"_"+str(self.halo_number)

    @property
    def path(self):
        return self.timestep.path+"/"+self.basename

    @property
    def handler(self):
        if not hasattr(self, "_handler"):
            self._handler = self.timestep.simulation.get_output_handler()
        return self._handler


    def load(self, mode=None):
        """Use pynbody to load the data for this halo, if it is present on this computer's filesystem.

        By default, the entire simulation is loaded and a subview with this halo is returned. If mode='partial',
        pynbody's partial loading system is used to only load the data for the one halo, saving memory."""

        return self.handler.load_object(self.timestep.extension, self.finder_id, object_typetag=self.tag, mode=mode)

    def calculate(self, name, return_description=False):
        """Use the live-calculation system to calculate a user-specified function of the stored data.

        See live_calculation.md for an introduction to this powerful functionality."""
        from .. import live_calculation
        calculation = live_calculation.parser.parse_property_name_if_required(name)
        (value,), description = calculation.values_sanitized_and_description([self])
        if len(value)==1:
            retval = value[0]
        else:
            retval = value

        if return_description:
            return retval, description
        else:
            return retval

    def __getitem__(self, key):
        """Highest-level method for retrieving data or link"""
        return self.get_data(key)

    def get(self, key, default=None):
        """Highest-level method for retrieving data or link, with default return value if no data is found
        under the specified key"""
        try:
            return self.get_data(key)
        except KeyError:
            return default

    def get_data(self, key, raw=False, always_return_array=False):
        """High-level data access to the property or linked halo named by key.

        :param key: string with the name of the property or link
        :param raw: if True, get the raw data rather than attempt to reassemble the data into a science-ready form
        :param always_return_array: if True, always return the data as a list of values, even if there is only one
           stored property with the specified name
        """

        if raw:
            getters=[extraction_patterns.halo_property_raw_value_getter]
        else:
            getters=[extraction_patterns.halo_property_value_getter]
        getters+=[extraction_patterns.halo_link_target_getter]

        return_data = self.get_objects(key, getters)

        if (not always_return_array) and len(return_data) == 1:
            return_data = return_data[0]

        return return_data

    def get_objects(self, key, getters = [extraction_patterns.halo_property_getter,
                                          extraction_patterns.halo_link_getter]):
        """Get objects belonging to this halo named by the specified key.

        Compared to get_data, this allows access to the underlying HaloProperty or HaloLink objects, or to perform
        custom processing of the data by specifying particular extraction patterns to getters. For more information,
        see halo_data_extraction_patterns."""
        from . import Session
        session = Session.object_session(self)
        key_id = get_dict_id(key, session=session)

        ret_values = []
        for g in getters:
            ret_values += g.get(self, key_id, session)

        if len(ret_values) == 0:
            raise KeyError("No such property %r" % key)
        return ret_values


    def __setitem__(self, key, obj):
        if isinstance(obj, Halo):
            self._setitem_one_halo(key, obj)
        elif hasattr(obj, '__len__') and all([isinstance(x,Halo) for x in obj]):
            self._setitem_multiple_halos(key, obj)
        else:
            self._setitem_property(key, obj)

    def _setitem_property(self, key, obj):
        from . import Session
        from .halo_data import HaloProperty

        session = Session.object_session(self)
        key = get_or_create_dictionary_item(session, key)
        X = self.properties.filter_by(name_id=key.id).first()
        if X is not None:
            X.data = obj
        else:
            X = session.merge(HaloProperty(self, key, obj))
        X.creator_id = creator.get_creator_id()

    def _setitem_one_halo(self, key, obj):
        from . import Session, HaloLink
        session = Session.object_session(self)
        key = get_or_create_dictionary_item(session, key)
        X = self.links.filter_by(halo_from_id=self.id, relation_id=key.id).first()
        if X is None:
            X = session.merge(HaloLink(self, obj, key))
            X.creator_id = creator.get_creator_id()
        else:
            X.halo_to = obj

    def _setitem_multiple_halos(self, key, obj):
        from . import Session
        from .halo_data import HaloLink

        session = Session.object_session(self)
        key = get_or_create_dictionary_item(session, key)
        self.links.filter_by(halo_from_id=self.id, relation_id=key.id).delete()
        links = [HaloLink(self, halo_to, key) for halo_to in obj]
        session.add_all(links)


    def keys(self, getters = [extraction_patterns.halo_property_getter,
                              extraction_patterns.halo_link_getter]):
        from . import Session
        names = []
        session = Session.object_session(self)
        for g in getters:
            names+=g.keys(self, session)


        return names

    def __contains__(self, item):
        return item in list(self.keys())

    @property
    def earliest(self):
        if self.previous is not None:
            return self.previous.earliest
        else:
            return self

    @property
    def latest(self):
        if self.next is not None:
            return self.next.latest
        else:
            return self

    def plot(self, name, *args, **kwargs):
        from . import Session
        name_id = get_dict_id(name, Session.object_session(self))
        data = self.properties.filter_by(name_id=name_id).first()
        return data.plot(*args, **kwargs)


    def calculate_for_descendants(self, *plist, **kwargs):
        """Run the specified calculations on this halo and its descendants

        Each argument is a string (or an instance of live_calculation.Calculation), following the syntax
        described in live_calculation.md.

        *kwargs*:

        :param nmax: The maximum number of descendants to consider (default 1000)
        :param strategy: The class to use to find the descendants (default relation_finding.MultiHopMajorDescendantsStrategy)
        """
        from .. import live_calculation
        from .. import relation_finding
        from .. import temporary_halolist as thl
        from . import Session
        from .. import query as db_query

        nmax = kwargs.get('nmax',1000)
        strategy = kwargs.get('strategy', relation_finding.MultiHopMajorDescendantsStrategy)
        strategy_kwargs = kwargs.get('strategy_kwargs', {})

        if isinstance(plist[0], live_calculation.Calculation):
            property_description = plist[0]
        else:
            property_description = live_calculation.parser.parse_property_names(*plist)

        # must be performed in its own session as we intentionally load in a lot of
        # objects with incomplete lazy-loaded properties
        session = Session()
        try:
            with strategy(db_query.get_halo(self.id, session), nhops_max=nmax,
                          include_startpoint=True, **strategy_kwargs).temp_table() as tt:
                raw_query = thl.halo_query(tt)
                query = property_description.supplement_halo_query(raw_query)
                results = query.all()
                return property_description.values_sanitized(results)
        finally:
            session.close()

    def calculate_for_progenitors(self, *plist, **kwargs):
        """Run the specified calculations on the progenitors of this halo

        For more information see calculate_for_descendants.
        """
        from .. import relation_finding
        kwargs['strategy'] = relation_finding.MultiHopMajorProgenitorsStrategy
        return self.calculate_for_descendants(*plist, **kwargs)

    def reverse_property_cascade(self, *args, **kwargs):
        """The old alias for calculate_for_progenitors, retained for compatibility"""
        return self.calculate_for_progenitors(*args, **kwargs)

    def property_cascade(self, *args, **kwargs):
        """The old alias for calculate_for_descendants, retained for compatibility"""
        return self.calculate_for_descendants(*args, **kwargs)


    @property
    def next(self):
        if not hasattr(self, '_next'):
            from .. import relation_finding
            strategy = relation_finding.HopMajorDescendantStrategy(self)
            self._next=strategy.first()

        return self._next

    @property
    def previous(self):
        if not hasattr(self, '_previous'):
            from .. import relation_finding
            strategy = relation_finding.HopMajorProgenitorStrategy(self)
            self._previous=strategy.first()

        return self._previous

    def short(self):
        return "<Halo " + str(self.halo_number) + " of ...>"

TimeStep.halos = orm.relationship(Halo, cascade='all', lazy='dynamic',
                                  primaryjoin=((Halo.timestep_id==TimeStep.id) & (Halo.object_typecode==0)),
                                  foreign_keys=Halo.timestep_id,
                                  order_by=Halo.halo_number)

class Tracker(Halo):
    __mapper_args__ = {
        'polymorphic_identity':3
    }

    tag = "tracker"

    def __init__(self, timestep, halo_number):
        super(Tracker, self).__init__(timestep, halo_number, halo_number, 0,0,0,
                                 self.__mapper_args__['polymorphic_identity'])

    @property
    def tracker(self):
        return self.timestep.simulation.trackers.filter_by(halo_number=self.halo_number).first()

    def load(self, mode=None):
        handler = self.timestep.simulation.get_output_handler()
        return handler.load_tracked_region(self.timestep.extension, self.tracker, mode=mode)



TimeStep.trackers = orm.relationship(Tracker, cascade='all', lazy='dynamic',
                                     primaryjoin=Tracker.timestep_id==TimeStep.id,
                                     order_by=Tracker.halo_number)




class BH(Tracker):
    __mapper_args__ = {
        'polymorphic_identity': 1
    }

    tag = "BH"

TimeStep.bhs = orm.relationship(BH, cascade='all', lazy='dynamic', primaryjoin=BH.timestep_id==TimeStep.id)


class Group(Halo):
    __mapper_args__ = {
        'polymorphic_identity':2
    }

    tag = "group"

    def __init__(self, *args):
        super(Group, self).__init__(*args)
        self.object_typecode = 2


TimeStep.groups = orm.relationship(Group, cascade='all', lazy='dynamic', primaryjoin=Group.timestep_id==TimeStep.id,
                                   order_by=Group.halo_number)





from __future__ import absolute_import
from .. import core
from . import message
from . import remote_import
import sys

class MessageRequestCreatorId(message.Message):
    def process(self):
        creator_id = core.creator.get_creator_id()
        MessageDeliverCreatorId(creator_id).send(self.source)

class MessageDeliverCreatorId(message.Message):
    pass


def synchronize_creator_object(session=None):
    """Causes the tangos module to use the rank 0 processor's 'Creator' object.

    Without calling this, each process of a distributed calculation has its own Creator object
    and so the user will see multiple calculations in their history when in effect they only ran
    one (through mpirun or similar)."""
    from . import parallel_backend_loaded

    if session is None:
        session = core.get_default_session()

    if not parallel_backend_loaded():
        return

    remote_import.ImportRequestMessage(__name__).send(0)
    MessageRequestCreatorId().send(0)
    id = MessageDeliverCreatorId.receive(0).contents
    core.creator.set_creator(session.query(core.creator.Creator).filter_by(id=id).first())



from __future__ import absolute_import
from __future__ import print_function
import glob
import os, os.path
from .. import config
from . import SimulationOutputSetHandler
from six.moves import range

class DummyTimestepData(object):
    def __init__(self, message, time, max_halos, halo=None):
        self.message = message
        self.halo = halo
        self.max_halos = max_halos
        self.time = time

    def __str__(self):
        return self.message

class TestOutputSetHandler(SimulationOutputSetHandler):
    def get_properties(self):
        result = {}
        with open(os.path.join(config.base, self.basename, "sim_info"),'r') as f:
            for line in f:
                line_split = line.split()
                result[line_split[0]] = " ".join(line_split[1:])
        return result

    def enumerate_timestep_extensions(self):
        pre_extension_length = len(os.path.join(config.base, self.basename))
        steps = glob.glob(os.path.join(config.base, self.basename, "step.*"))
        for i in steps:
            print(i, i[pre_extension_length:], self.strip_slashes(i[pre_extension_length:]))
            yield self.strip_slashes(i[pre_extension_length:])

    def get_timestep_properties(self, ts_extension):
        result = {}
        result['time_gyr'] = float(self._get_ts_property(ts_extension, 'time'))
        result['redshift'] = float(self._get_ts_property(ts_extension, 'redshift'))
        return result

    def enumerate_objects(self, ts_extension, object_typetag='halo', min_halo_particles=config.min_halo_particles):
        nhalos_string = self._get_ts_property(ts_extension, object_typetag+"s")
        nhalos = 0 if nhalos_string is None else int(nhalos_string)
        for i in range(nhalos):
            yield i+1, 2000-i, 0, 0

    def match_objects(self, ts1, ts2, halo_min, halo_max, dm_only=False, threshold=0.005, object_typetag='halo'):
        """Test implementation of match halos always links halo i->i, and a 0.05 mass transfer from i->i+1"""
        assert object_typetag=='halo' # currently only handle halos
        f1 = self.load_timestep(ts1)
        f2 = self.load_timestep(ts2)
        if halo_max is None:
            halo_max = f1.max_halos
        halo_max = min((halo_max,f1.max_halos,f2.max_halos))
        return_matches = [tuple()]
        for i in range(1,halo_max+1):
            if i>=halo_min:
                return_matches.append(((i, 1.0),(i+1,0.05),))
            else:
                return_matches.append(tuple())
        return return_matches

    def load_timestep(self, ts_extension, mode=None):
        return DummyTimestepData("Test string - this would contain the data for "+ts_extension,
                                 float(self._get_ts_property(ts_extension, 'time')),
                                 int(self._get_ts_property(ts_extension, 'halos')))

    def load_region(self, ts_extension, region_specification, mode=None):
        data = self.load_timestep(ts_extension)
        data.message = data.message[region_specification]
        return data

    def load_object(self, ts_extension, halo_number, object_typetag='halo', mode=None):
        assert object_typetag=='halo'
        return DummyTimestepData("Test string - this would contain the data for %s halo %d"%(ts_extension ,halo_number),
                                 float(self._get_ts_property(ts_extension, 'time')),
                                 int(self._get_ts_property(ts_extension, 'halos')),
                                 halo_number)

    def _get_ts_property(self, ts_extension, property):
        ts_filename = self._extension_to_filename(ts_extension)
        with open(ts_filename, 'r') as f:
            for line in f:
                line_split = line.split()
                if line_split[0].lower() == property.lower():
                    return " ".join(line_split[1:])



from __future__ import absolute_import
from .. import BuiltinFunction, FixedNumericInput
import numpy as np
import functools
from six.moves import zip

@BuiltinFunction.register
def abs(halos, vals):
    return arithmetic_unary_op(vals, functools.partial(np.linalg.norm, axis=-1))

@BuiltinFunction.register
def sqrt(halos, vals):
    return arithmetic_unary_op(vals, np.sqrt)

@BuiltinFunction.register
def log(halos, vals):
    return arithmetic_unary_op(vals, np.log)

@BuiltinFunction.register
def log10(halos, vals):
    return arithmetic_unary_op(vals, np.log10)

@BuiltinFunction.register
def subtract(halos, vals1, vals2):
    return arithmetic_binary_op(vals1, vals2, np.subtract)

@BuiltinFunction.register
def add(halos, vals1, vals2):
    return arithmetic_binary_op(vals1, vals2, np.add)

@BuiltinFunction.register
def divide(halos, vals1, vals2):
    return arithmetic_binary_op(vals1, vals2, np.divide)

@BuiltinFunction.register
def multiply(halos, vals1, vals2):
    return arithmetic_binary_op(vals1, vals2, np.multiply)

@BuiltinFunction.register
def greater(halos, vals1, vals2):
    return arithmetic_binary_op(vals1, vals2, np.greater)

@BuiltinFunction.register
def less(halos, vals1, vals2):
    return arithmetic_binary_op(vals1, vals2, np.less)

@BuiltinFunction.register
def logical_and(halos, vals1, vals2):
    return arithmetic_binary_op(vals1, vals2, np.logical_and)

@BuiltinFunction.register
def logical_or(halos, vals1, vals2):
    return arithmetic_binary_op(vals1, vals2, np.logical_or)

@BuiltinFunction.register
def logical_not(halos, vals):
    return arithmetic_unary_op(vals, np.logical_not)

@BuiltinFunction.register
def power(halos, vals1, vals2):
    return arithmetic_binary_op(vals1, vals2, np.power)

def arithmetic_binary_op(vals1, vals2, op):
    results = []
    for v1,v2 in zip(vals1, vals2):
        if v1 is not None and v2 is not None:
            v1 = np.asarray(v1, dtype=float)
            v2 = np.asarray(v2, dtype=float)
            result = op(v1,v2)
        else:
            result = None
        results.append(result)
    return results

def arithmetic_unary_op(vals1, op):
    results = []
    for v1 in vals1:
        if v1 is not None:
            v1 = np.asarray(v1, dtype=float)
            result = op(v1)
        else:
            result = None
        results.append(result)
    return results


from __future__ import absolute_import

yt = None # deferred import; occurs when a YtOutputSetHandler is constructed

from . import finding
from . import SimulationOutputSetHandler
from .. import config
from ..log import logger
from six.moves import range

class YtOutputSetHandler(finding.PatternBasedFileDiscovery, SimulationOutputSetHandler):
    def __init__(self, *args, **kwargs):
        super(YtOutputSetHandler, self).__init__(*args, **kwargs)
        global yt
        import yt as yt_local
        yt = yt_local
        assert "3.4.0" <= yt.__version__, "Tangos requires yt version 3.4.0 or later"

    def get_timestep_properties(self, ts_extension):
        ts_filename =  self._extension_to_filename(ts_extension)
        f = yt.load(ts_filename)
        time_gyr = float(f.current_time.in_units("Gyr"))
        redshift = f.current_redshift
        results = {'time_gyr': time_gyr, 'redshift': redshift,
                   'available': True}
        return results

    def load_timestep_without_caching(self, ts_extension, mode=None):
        if mode!=None:
            raise ValueError("Custom load modes are not supported with yt")
        f = yt.load(self._extension_to_filename(ts_extension))
        return f

    def load_object(self, ts_extension, object_number, object_typetag='halo', mode=None):
        f = self.load_timestep(ts_extension, mode)
        cat, cat_dat = self._load_halo_cat(ts_extension, object_typetag)
        center = cat_dat["halos","particle_position"][object_number]
        center+=f.domain_left_edge-cat.domain_left_edge
        radius = cat_dat["halos","virial_radius"][object_number]
        return f.sphere(center.in_cgs(), radius.in_cgs())


    def load_tracked_region(self, ts_extension, track_data, mode=None):
        raise NotImplementedError("Tracked regions not implemented for yt")

    def match_objects(self, ts1, ts2, halo_min, halo_max, dm_only=False, threshold=0.005, object_typetag='halo'):
        raise NotImplementedError("Matching halos still needs to be implemented for yt")

    def enumerate_objects(self, ts_extension, object_typetag="halo", min_halo_particles=config.min_halo_particles):
        if object_typetag!="halo":
            raise StopIteration
        if self._can_enumerate_objects_from_statfile(ts_extension, object_typetag):
            for X in self._enumerate_objects_from_statfile(ts_extension, object_typetag):
                yield X
        else:
            logger.warn("No halo statistics file found for timestep %r", ts_extension)
            logger.warn(" => enumerating %ss directly using yt", object_typetag)

            catalogue, catalogue_data = self._load_halo_cat(ts_extension, object_typetag)
            num_objects = len(catalogue_data["halos", "virial_radius"])

            for i in range(num_objects):
                obj = self.load_object(ts_extension, i, object_typetag)
                NDM = len(obj["DarkMatter","Mass"])
                NGas = len(obj["Gas","Mass"])
                NStar = len(obj["Stars","Mass"])
                if NDM > min_halo_particles:
                    yield i, NDM, NStar, NGas


    def _load_halo_cat(self, ts_extension, object_typetag):
        snapshot_file = self.load_timestep(ts_extension)
        if object_typetag== 'halo':
            if not hasattr(snapshot_file, "_tangos_halos"):
                snapshot_file._tangos_halos = self._load_halo_cat_without_caching(ts_extension, snapshot_file)
            return snapshot_file._tangos_halos
        else:
            raise ValueError("Unknown halo type %r" % object_typetag)

    def _load_halo_cat_without_caching(self, ts_extension, snapshot_file):
        raise NotImplementedError("You need to select a subclass of YtOutputSetHandler")

    def get_properties(self):
        return {}


class YtChangaAHFOutputSetHandler(YtOutputSetHandler):
    patterns = ["*.00???", "*.00????"]

    def _load_halo_cat_without_caching(self, ts_extension, snapshot_file):
        cat = yt.frontends.ahf.AHFHalosDataset(self._extension_to_filename("halos/"+ts_extension)+".AHF_param",
                                                hubble_constant = snapshot_file.hubble_constant)
        cat_data = cat.all_data()
        return cat, cat_data




from __future__ import absolute_import

from __future__ import print_function
import multiprocessing
import threading
import warnings
import time
import os
import signal
from six.moves import range
from six.moves import zip

_slave = False
_rank = None
_size = None
_pipe = None
_recv_lock = None
_recv_buffer = []

_print_exceptions = True

class NoMatchingItem(Exception):
    pass


def send(data, destination, tag=0):
    _pipe.send((data, destination, tag))

def receive_any(source=None):
    return receive(source,None,True)


def receive(source=None, tag=0, return_tag=False):
    while True:
        try:
            item = _pop_first_match_from_reception_buffer(source, tag)
            if return_tag:
                return item
            else:
                return item[0]
        except NoMatchingItem:
            _receive_item_into_buffer()


NUMPY_SPECIAL_TAG = 1515

def send_numpy_array(data, destination):
    send(data,destination,tag=NUMPY_SPECIAL_TAG)

def receive_numpy_array(source):
    return receive(source,tag=NUMPY_SPECIAL_TAG)

def _pop_first_match_from_reception_buffer(source, tag):
    for item in _recv_buffer:
        if ((item[2] == tag or tag is None) and (item[1] == source or source is None)):
            # consume item
            _recv_buffer.remove(item)
            return item

    raise NoMatchingItem()

def _receive_item_into_buffer():
    if _recv_lock.acquire(False):
        try:
            _recv_buffer.append(_pipe.recv())
        finally:
            _recv_lock.release()
    else:
        # block until a data item has been received by another thread
        _recv_lock.acquire()
        _recv_lock.release()



def rank():
    return _rank

def size():
    return _size

def barrier():
    pass

def finalize():
    _pipe.send("finalize")

def launch_wrapper(target_fn, rank_in, size_in, pipe_in, args_in):
    global _slave, _rank, _size, _pipe, _recv_lock
    _rank = rank_in
    _size = size_in
    _pipe = pipe_in
    _recv_lock = threading.Lock()
    try:
        target_fn(*args_in)
        finalize()
    except Exception as e:
        import sys, traceback
        exc_type, exc_value, exc_traceback = sys.exc_info()
        global _print_exceptions
        if _print_exceptions:
            print("Error on a sub-process:", file=sys.stderr)
            traceback.print_exception(exc_type, exc_value, exc_traceback,
                                      file=sys.stderr)
        _pipe.send(("error", e))

    _pipe.close()


def launch_functions(functions, args):
    global _slave
    if _slave:
        raise RuntimeError("Multiprocessing session is already underway")

    num_procs = len(functions)

    child_connections, parent_connections = list(zip(*[multiprocessing.Pipe() for rank in range(num_procs)]))
    processes = [multiprocessing.Process(target=launch_wrapper, args=(function, rank, num_procs, pipe, args_i))
                 for rank, (pipe, function, args_i) in
                 enumerate(zip(child_connections, functions, args))]

    for proc_i in processes:
        proc_i.start()

    running = [True for rank in range(num_procs)]
    error = False

    while any(running):
        for i, pipe_i in enumerate(parent_connections):
            if pipe_i.poll():
                message = pipe_i.recv()
                if message=='finalize':
                    #print "  ---> multiprocessing backend: finalize node ",i,running
                    running[i]=False
                elif message[0]=='error':
                    error = message[1]
                    running = [False]
                    break
                else:
                    #print "multiprocessing backend: pass message ",i,"->",message[1]
                    parent_connections[message[1]].send((message[0],i,message[2]))

    #print "multiprocessing backend: all finished"

    for pipe_i in parent_connections:
        pipe_i.close()

    for proc_i in processes:
        if error:
            #print "multiprocessing backend: send signal to",proc_i.pid
            os.kill(proc_i.pid, signal.SIGTERM)
        proc_i.join()

    if error:
        raise error



def launch(function, num_procs, args):
    if num_procs is None:
        raise RuntimeError("To launch a parallel session using multiprocessing backend, you need to specify the number of processors")

    launch_functions([function]*num_procs, [args]*num_procs)





from __future__ import absolute_import
from __future__ import print_function
from sqlalchemy import and_

from tangos import get_default_session, Creator, Base
from tangos.core import Simulation, TimeStep, Halo, HaloProperty
from six.moves import map
import six


def all_simulations(session=None):
    if session is None:
        session = get_default_session()
    return session.query(Simulation).all()


def all_creators():
    return get_default_session().query(Creator).all()


def get_simulation(id, session=None):
    if session is None:
        session = get_default_session()
    if isinstance(id, str) or isinstance(id, six.text_type):
        assert "/" not in id
        if "%" in id:
            match_clause = Simulation.basename.like(id)
        else:
            match_clause = Simulation.basename == id
        res = session.query(Simulation).filter(match_clause)
        num = res.count()
        if num == 0:
            raise RuntimeError("No simulation matches %r" % id)
        elif num > 1:
            raise RuntimeError("Multiple (%d) matches for %r" % (num, id))
        else:
            return res.first()

    else:
        return session.query(Simulation).filter_by(id=int(id)).first()


def get_timestep(id, session=None, sim=None):
    if session is None:
        session = get_default_session()
    if isinstance(id, str) or isinstance(id, six.text_type):
        if sim is None:
            sim, ts = id.split("/")
            sim = get_simulation(sim)
        else:
            ts = id
        res = session.query(TimeStep).filter(
            and_(TimeStep.extension.like(ts), TimeStep.simulation_id == sim.id))
        num = res.count()
        if num == 0:
            raise RuntimeError("No timestep matches for %r" % id)
        elif num > 1:
            raise RuntimeError("Multiple (%d) matches for timestep %r of simulation %r" % (
                num, ts, sim))
        else:
            return res.first()
    else:
        return session.query(TimeStep).filter_by(id=int(id)).first()


def get_halo(id, session=None):
    """Get a halo from an ID or an identifying string

    Optionally, use the specified session.

    :rtype: Halo
    """
    if session is None:
        session = get_default_session()

    if isinstance(id, str):
        sim, ts, halo = id.split("/")
        ts = get_timestep(sim + "/" + ts)
        return ts[halo]
    else:
        return session.query(Halo).filter_by(id=int(id)).first()


def get_item(path, session=None):
    c = path.count("/")
    if c is 0:
        return get_simulation(path, session)
    elif c is 1:
        return get_timestep(path, session)
    elif c is 2:
        return get_halo(path, session)


def get_haloproperty(id):
    return get_default_session().query(HaloProperty).filter_by(id=id).first()


def get_items(path_list, session=None):
    return [get_item(path,session) for path in path_list]


def getdb(cl) :
    """Function decorator to ensure input is parsed into a database object."""
    def getdb_inner(f) :
        def wrapped(*args, **kwargs) :

            if not isinstance(args[0], Base) :
                args = list(args)
                if isinstance(args[0], int) :
                    item = get_default_session().query(cl).filter_by(id=args[0]).first()
                else :
                    item = get_item(args[0])
                if not isinstance(item, cl) :
                    if isinstance(item, Simulation) and cl is Halo:
                        print("Picking first timestep and first halo")
                        item = item.timesteps[0].halos[0]
                    else :
                        raise RuntimeError("Path points to wrong type of db object %r"%item)
                args[0] = item
            return f(*args,**kwargs)
        return wrapped
    return getdb_inner


__all__ = ['all_simulations', 'all_creators', 'get_simulation', 'get_timestep',
           'get_halo', 'get_item' ,'get_haloproperty', 'get_items', 'getdb']


import numpy as np
from .spherical_region import SphericalRegionHaloProperties
from .centring import centred_calculation

class HaloDensityProfile(SphericalRegionHaloProperties):
    # include

    names = "dm_density_profile", "dm_mass_profile"

    def plot_x0(self):
        return self.plot_xdelta()/2

    def plot_xdelta(self):
        return self._simulation.get("approx_resolution_kpc", 0.1)

    def plot_xlabel(self):
        return "r/kpc"

    def plot_ylabel(self):
        return r"$\rho/M_{\odot}\,kpc^{-3}$", r"$M/M_{\odot}$"

    def _get_profile(self, halo, maxrad):
        import pynbody

        delta = self.plot_xdelta()
        nbins = int(maxrad / delta)
        maxrad = delta * (nbins + 1)

        pro = pynbody.analysis.profile.Profile(halo, type='lin', ndim=3,
                                               min=0, max=maxrad, nbins=nbins)

        rho_a = pro['density']
        mass_a = pro['mass_enc']

        rho_a = rho_a.view(np.ndarray)
        mass_a = mass_a.view(np.ndarray)

        return rho_a, mass_a

    @centred_calculation
    def calculate(self, data, existing_properties):

        dm_a, dm_b = self._get_profile(data.dm, existing_properties["max_radius"])

        return dm_a, dm_b


class BaryonicHaloDensityProfile(HaloDensityProfile):
    names = "gas_density_profile", "gas_mass_profile", "star_density_profile", "star_mass_profile"

    @centred_calculation
    def calculate(self, data, existing_properties):
        gas_a, gas_b = self._get_profile(data.gas, existing_properties["max_radius"])
        star_a, star_b = self._get_profile(data.star, existing_properties["max_radius"])
        return gas_a, gas_b, star_a, star_b



from __future__ import absolute_import
from __future__ import print_function
from sqlalchemy import Column, Integer, String, DateTime

from . import Base, get_default_session

_current_creator = None

class Creator(Base):
    __tablename__ = 'creators'

    id = Column(Integer, primary_key=True)
    command_line = Column(String)
    dtime = Column(DateTime)
    host = Column(String)
    username = Column(String)
    cwd = Column(String)

    def __repr__(self):
        return "<Creator " + self.username + " on " + self.host + " @ " + self.dtime.strftime("%d/%m/%y %H:%M") + " via " + self.command_line.split(" ")[0].split("/")[-1] + ">"

    def __init__(self, argv=None):
        import socket
        import getpass
        import datetime
        import os
        if argv == None:
            import sys
            argv = sys.argv

        self.command_line = " ".join(argv)
        self.host = socket.gethostname()
        self.username = getpass.getuser()
        self.dtime = datetime.datetime.now()
        self.cwd = os.getcwd()

    def print_info(self):
        run = self
        print("*" * 60)
        print("Run ID = ", self.id)
        print("Command line = ", self.command_line)
        print("Host = ", self.host)
        print("Username = ", self.username)
        print("Time = ", self.dtime.strftime("%d/%m/%y %H:%M"))
        if len(run.simulations) > 0:
            print(">>>   ", len(run.simulations), "simulations")
        if run.timesteps.count() > 0:
            print(">>>   ", (run.timesteps).count(), "timesteps")
        if run.halos.count() > 0:
            print(">>>   ", run.halos.count(), "halos")
        if run.halolinks.count() > 0:
            print(">>>   ", run.halolinks.count(), "halolinks")
        if run.properties.count() > 0:
            print(">>>   ", run.properties.count(), "halo properties")
        if run.simproperties.count() > 0:
            print(">>>   ", run.simproperties.count(), "simulation properties")


def get_creator(session=None):
    """Get a Creator object for this process, for the specified session.

    If session is None, return the object for the default session."""
    from sqlalchemy import inspect
    global _current_creator

    if session is None:
        session = get_default_session()

    if _current_creator is None:
        _current_creator = get_default_session().merge(Creator())
    if not inspect(_current_creator).persistent:
        get_default_session().commit()
    assert inspect(_current_creator).persistent
    
    if session is get_default_session():
        return _current_creator
    else:
        return session.query(Creator).filter_by(id=_current_creator.id).first()


def get_creator_id():
    return get_creator().id

def set_creator(creator):
    """Set the Creator object to be used in all writes during the lifetime of the current process.

    A Creator object is normally constructed automatically, but this function allows all future writes to be
    associated with a different Creator. This is mainly used by MPI runs to give the illusion that all data was
    written by one process."""
    global _current_creator
    _current_creator = creator





from __future__ import absolute_import
from __future__ import print_function
import tangos as db
import sqlalchemy

import tangos.core.dictionary
import tangos.core.halo
import tangos.core.halo_data


def fix_reverse_links(session, name_forwards, name_reverse):
    """Remove any existing links with given name_reverse, and regenerate them from the links called name_forwards
    """
    dict_forwards = tangos.core.dictionary.get_or_create_dictionary_item(db.core.get_default_session(), name_forwards)
    dict_reverse = tangos.core.dictionary.get_or_create_dictionary_item(db.core.get_default_session(), name_reverse)
    session.query(tangos.core.halo_data.HaloLink).filter_by(relation_id=dict_reverse.id).delete()
    session.commit()
    connection = session.connection()

    query = session.query(tangos.core.halo_data.HaloLink.halo_to_id, tangos.core.halo_data.HaloLink.halo_from_id,
                          sqlalchemy.literal(dict_reverse.id), tangos.core.halo_data.HaloLink.weight).filter_by(relation_id=dict_forwards.id)
    insert = tangos.core.halo_data.HaloLink.__table__.insert().\
                       from_select(['halo_from_id','halo_to_id','relation_id','weight'],
                                   query)
    print(insert)
    connection.execute(insert)
    session.commit()

def main():
    fix_reverse_links(db.core.get_default_session(),"BH_central","host_halo")



from __future__ import absolute_import
import tangos as db
import tangos.relation_finding
import numpy as np


def get_mergers_of_major_progenitor(input_halo):
    """Given a halo, return the redshifts and ratios of all mergers on the major progenitor branch

    :parameter input_halo - the halo to consider the history of
    :type input_halo tangos.core.Halo

    :returns redshift, ratio, halo - arrays of the redshifts, ratios (1:X) and halo DB objects for the mergers
    """
    redshift = []
    ratio = []
    halo  = []
    while input_halo is not None:
        mergers = db.relation_finding.MultiHopMostRecentMergerStrategy(input_halo).all()
        if len(mergers)>0 :
            for m in mergers[1:]:
                redshift.append(mergers[0].timestep.next.redshift)
                halo.append((mergers[0], m))
                ratio.append(float(mergers[0].NDM)/m.NDM)
            input_halo = mergers[0]
        else:
            input_halo = None

    return np.array(redshift), np.array(ratio), halo

def most_major_mergers_since(ts, Mvir_min=0.8e12, Mvir_max=2e12, z_merger_max=4.0, no_merger_value=None):
    """Given a timestep, calculate the most major merger ratio since a given redshift for all halos in a specified mass range

    :parameter ts - The (end) timestep to take halos from
    :parameter Mvir_min, Mvir_max - the minimum and maximum halo masses to consider
    :parameter z_merger_max - the maximum redshift to search for mergers
    :parameter no_merger_value - the value to insert for a halo if no merger was found (default None)
    :type ts tangos.core.TimeStep

    """
    dbid, Mvir = ts.calculate_all("dbid()","Mvir")
    mask = (Mvir>Mvir_min)*(Mvir<Mvir_max)

    most_major = []
    for current_id in dbid[mask]:
        halo = db.get_halo(current_id)
        redshift, ratio, halos = get_mergers_of_major_progenitor(halo)
        m_mask = np.array(redshift)<z_merger_max
        if m_mask.sum()>0:
            most_major.append(min(ratio[m_mask]))
        else:
            most_major.append(no_merger_value)
    return np.array(most_major)



from __future__ import absolute_import

import warnings

def send(data, destination, tag=0):
    raise RuntimeError("Cannot send data to another CPU: parallelism is disabled")

def receive(source=None, tag=0):
    raise RuntimeError("Cannot receive data from another CPU: parallelism is disabled")

def receive_any(source=None):
    raise RuntimeError("Cannot receive data from another CPU: parallelism is disabled")

def rank():
    return 0

def size():
    return 1

def barrier():
    pass

def finalize():
    pass


def launch(function, num_procs, args):
    if num_procs is not None and num_procs!=1:
        warnings.warn("Number of processors requested (%d) will be ignored as this is a single-processor run"%num_procs)
    function(*args)


#!/usr/bin/env python

from __future__ import absolute_import
from __future__ import print_function
import argparse

import tangos as db
import tangos.core.timestep
from tangos.simulation_output_handlers import halo_stat_files as hsf
from tangos.util.terminalcontroller import term


def main():

    parser = argparse.ArgumentParser()
    db.supplement_argparser(parser)

    parser.add_argument('--sims','--for', action='store', nargs='*',
                                metavar='simulation_name',
                                help='Specify a simulation (or multiple simulations) to run on')

    parser.add_argument('properties', action='store', nargs='+',
                            help="The names of the halo properties to import from the AHF_halos file")

    parser.add_argument('--backwards', action='store_true',
                            help='Process low-z timesteps first')

    args = parser.parse_args()

    db.process_options(args)


    base_sim = db.sim_query_from_name_list(args.sims)

    names = args.properties

    for x in base_sim:
        timesteps = db.core.get_default_session().query(tangos.core.timestep.TimeStep).filter_by(
            simulation_id=x.id, available=True).order_by(tangos.core.timestep.TimeStep.redshift.desc()).all()

        if args.backwards:
            timesteps = timesteps[::-1]

        for ts in timesteps:
            print(term.GREEN, "Processing ",ts, term.NORMAL)
            hsf.HaloStatFile(ts).add_halo_properties(*names)
            db.core.get_default_session().commit()



#!/usr/bin/env python2.7

from __future__ import absolute_import
from __future__ import print_function
import tangos as db
import tangos.core.dictionary
import tangos.core.halo
import tangos.core.halo_data
import tangos.core.timestep
import tangos.core.tracking
import tangos.parallel_tasks as parallel_tasks
import tangos.parallel_tasks.database
import tangos.tracking
from tangos.log import logger
import numpy as np
import glob
import sys
import six
from six.moves import zip


def resolve_multiple_mergers(bh_map):
    for k,v in six.iteritems(bh_map):
        if v[0] in bh_map:
            old_target = v[0]
            old_weight = v[1]
            bh_map[k] = bh_map[old_target][0],v[1]*bh_map[old_target][1]
            print("Reassignment:",k,old_target,bh_map[k])
            resolve_multiple_mergers(bh_map)
            return

def generate_halolinks(session, fname, pairs):

    for ts1, ts2 in pairs:
        links = []
        mergers_links = []
        bh_map = {}
        logger.info("Gathering BH tracking information for steps %r and %r", ts1, ts2)
        with parallel_tasks.RLock("bh"):
            dict_obj = db.core.get_or_create_dictionary_item(session, "tracker")
            dict_obj_next = db.core.get_or_create_dictionary_item(session, "BH_merger_next")
            dict_obj_prev = db.core.get_or_create_dictionary_item(session, "BH_merger_prev")

        track_links_n, idf_n, idt_n = db.tracker.get_tracker_links(session, dict_obj_next)
        halos_1, nums1, id1 = db.tracker.get_tracker_halos(ts1)
        halos_2, nums2, id2 = db.tracker.get_tracker_halos(ts2)
        tracker_links, idf, idt = db.tracker.get_tracker_links(session,dict_obj)

        if len(nums1) == 0 or len(nums2) == 0:
            logger.info("No Tracker Halos found in either step %r or %r... moving on", ts1, ts2)
            continue

        logger.info("Collecting BH tracker links between steps %r and %r", ts1, ts2)
        o1 = np.where(np.in1d(nums1,nums2))[0]
        o2 = np.where(np.in1d(nums2,nums1))[0]
        if len(o1) == 0 or len(o2) == 0:
            continue
        with session.no_autoflush:
            for ii, jj in zip(o1,o2):
                if nums1[ii] != nums2[jj]:
                    raise RuntimeError("ERROR mismatch of BH iords")
                exists = np.where((idf==id1[ii])&(idt==id2[jj]))[0]
                if len(exists) == 0:
                    links.append(tangos.core.halo_data.HaloLink(halos_1[ii],halos_2[jj],dict_obj,1.0))
                    links.append(tangos.core.halo_data.HaloLink(halos_2[jj],halos_1[ii],dict_obj,1.0))
        logger.info("Found %d tracker links between steps %r and %r", len(links), ts1, ts2)

        logger.info("Gathering BH Merger information for steps %r and %r", ts1, ts2)
        for l in open(fname[0]):
            l_split = l.split()
            t = float(l_split[0])
            bh_dest_id = int(l_split[2])
            bh_src_id = int(l_split[3])
            ratio = float(l_split[4])

            if t>ts1.time_gyr and t<=ts2.time_gyr:
                bh_map[bh_src_id] = (bh_dest_id, ratio)

        resolve_multiple_mergers(bh_map)
        logger.info("Gathering BH Merger links for steps %r and %r", ts1, ts2)
        with session.no_autoflush:
            for src,(dest,ratio) in six.iteritems(bh_map):
                ob = np.where(nums1 == src)[0]
                oa = np.where(nums2 == dest)[0]
                if len(oa) == 0 or len(ob) == 0:
                    continue
                bh_src_before = halos_1[ob[0]]
                bh_dest_after = halos_2[oa[0]]

                if len(np.where((idf_n==id1[ob[0]])&(idt_n==id2[oa[0]]))[0]) == 0:
                    mergers_links.append(tangos.core.halo_data.HaloLink(bh_src_before,bh_dest_after,dict_obj_next,1.0))
                    mergers_links.append(tangos.core.halo_data.HaloLink(bh_dest_after,bh_src_before,dict_obj_prev,ratio))
        logger.info("Found %d BH Merger links for steps %r and %r", len(mergers_links), ts1, ts2)

        with parallel_tasks.RLock("bh"):
            logger.info("Committing total %d BH links for steps %r and %r", len(mergers_links)+len(links), ts1, ts2)
            session.add_all(links)
            session.add_all(mergers_links)
            session.commit()
            logger.info("Finished Committing BH links for steps %r and %r", ts1, ts2)

def run():
    parallel_tasks.database.synchronize_creator_object()
    session = db.core.get_default_session()
    query = db.sim_query_from_args(sys.argv, session)
    for sim in query.all():
        pairs = parallel_tasks.distributed(list(zip(sim.timesteps[:-1],sim.timesteps[1:])))
        fname = glob.glob(db.config.base+"/"+sim.basename+"/*.mergers")
        if len(fname)==0:
            print("No merger file for "+sim.basename)
            return
        elif len(fname)>1:
            print("Can't work out which is the merger file for "+sim.basename)
            print("Found: ",fname)
            return
        with session.no_autoflush:
            generate_halolinks(session, fname, pairs)


def main():
    parallel_tasks.launch(run)


from __future__ import absolute_import
from sqlalchemy import Column, Integer, ForeignKey, Float, LargeBinary, Boolean
from sqlalchemy.orm import relationship, backref

from .. import Base
from .. import creator
from ..dictionary import DictionaryItem
from ..halo import Halo


class HaloLink(Base):
    __tablename__ = 'halolink'

    id = Column(Integer, primary_key=True)
    halo_from_id = Column(Integer, ForeignKey('halos.id'))
    halo_to_id = Column(Integer, ForeignKey('halos.id'))

    halo_from = relationship(Halo, primaryjoin=halo_from_id == Halo.id,
                             backref=backref('links', cascade='all',
                                             lazy='dynamic',
                                             primaryjoin=halo_from_id == Halo.id),
                             cascade='')

    halo_to = relationship(Halo, primaryjoin=(halo_to_id == Halo.id),
                           backref=backref('reverse_links', cascade='all, delete-orphan',
                                           lazy='dynamic',
                                           primaryjoin=halo_to_id == Halo.id),
                           cascade='')

    weight = Column(Float)

    creator_id = Column(Integer, ForeignKey('creators.id'))
    creator = relationship(creator.Creator, backref=backref(
        'halolinks', cascade='all, delete', lazy='dynamic'), cascade='save-update')

    relation_id = Column(Integer, ForeignKey('dictionary.id'))
    relation = relationship(DictionaryItem, primaryjoin=(
        relation_id == DictionaryItem.id), cascade='save-update,merge')



    def __init__(self,  halo_from, halo_to, relationship, weight=1.0):
        self.halo_from = halo_from
        self.halo_to = halo_to

        self.relation = relationship
        self.weight = weight

        self.creator_id = creator.get_creator_id()

    def __repr__(self):
        return "<HaloLink " + str(self.relation.text) + " " + str(self.halo_from.path) + " to " + str(self.halo_to.path) \
               + " weight=%.2f>"%self.weight




from . import PynbodyHaloProperties
from .. import LiveHaloProperties
import numpy as np
import functools
import contextlib

class CentreAndRadius(PynbodyHaloProperties):
    names = "shrink_center", "max_radius"

    def calculate(self, halo, existing_properties):
        # this does not appear at module level because we want tangos to be importable even when pynbody is not
        # present:
        import pynbody

        # ensure the box is wrapped correctly by centring on one of the particles:
        temporary_centre = np.array(halo['pos'][0])
        with _recenter(halo, temporary_centre):
            center = pynbody.analysis.halo.shrink_sphere_center(halo.dm, shrink_factor=0.8, velocity=False)

            # mark_timer can be used to track timing of parts of the calculation. The results of these timings
            # appears in the tangos_writer logs:
            self.mark_timer('cen')

            if any(center != center):
                raise RuntimeError("Something bizarre has happened with the centering")

            # measure rmax from the robust centre we have now identified:
            halo['pos'] -= center

            center+=temporary_centre  # centre should be relative to original snapshot!

            rmax = pynbody.derived.r(halo).max()
            # N.B. not using halo['r'] which could end up calculating r across entire simulation, needlessly

        return center.view(np.ndarray), rmax


class CentreAndRadiusComoving(LiveHaloProperties):
    names = "shrink_center_comoving", "max_radius_comoving"

    def calculate(self, _, existing_properties):
        return existing_properties['shrink_center']



@contextlib.contextmanager
def _recenter(halo, centre):
    original_positions = np.array(halo['pos'])  # take a copy so we can put everything back at the end
    halo['pos']-=centre
    halo.wrap()
    yield
    halo['pos']=original_positions


def centred_calculation(fn):
    """Wrap a calculation with a robust recentring of the halo particles that is automatically reverted"""
    @functools.wraps(fn)
    def new_fn(self, halo, existing_properties):
        with _recenter(halo, existing_properties['shrink_center']):
            return fn(self, halo, existing_properties)

    return new_fn



"""Configuration module for tangos

Rather than change anything directly in this file, you can create a config_local.py with the variable you
want to override and it will automatically take precedence.
"""

from __future__ import absolute_import
import os
import sys

home = os.environ['HOME']

if sys.platform=='darwin' :
    home+='/Science/'

db = os.environ.get("TANGOS_DB_CONNECTION", home+"/tangos_data.db")
base = os.environ.get("TANGOS_SIMULATION_FOLDER", home+"/")

default_fileset_handler_class = "pynbody.PynbodyOutputSetHandler"

num_multihops_max_default = 100
# the maximum number of links to follow when searching for related halos

default_linking_threshold = 0.005
# the percentage of particles in common between two objects before the database even bothers to store the relationship

min_halo_particles = 1000
# the minimum number of particles needed in an object before the database bothers to store it

default_backend = 'null'
# the default paralellism backend. Set e.g. to mpi4py to avoid having to pass --backend mpi4py to all parallel runs.


# names of property modules to import; default is for backwards compatibility on systems with N-Body-Shop extensions
property_modules = os.environ.get("TANGOS_PROPERTY_MODULES","tangos_nbodyshop_properties")
property_modules = property_modules.split(",")
property_modules = map(str.strip, property_modules)

file_ignore_pattern = []

max_traverse_depth = 3

# merger tree building thinning criteria
mergertree_min_fractional_weight = 0.02 # as a fraction of the weight of the strongest link from each halo
mergertree_min_fractional_NDM = 0 # as a fraction of the most massive halo at each timestep - set to zero for no thinning
mergertree_max_nhalos = 30 # maximum number of halos per step - discard the least massive ones
mergertree_timeout = 15.0 # seconds before abandoning the construction of a merger tree in the web interface


# On some network file systems, concurrency using sqlite is dodgy to say the least. After committing a transaction
# on one node, and before attempting to open a new transaction on another node, it seems empirically helpful to
# allow a significant time delay. This variable controls that delay.
DEFAULT_SLEEP_BEFORE_ALLOWING_NEXT_LOCK = 1.0
# number of seconds to sleep after a lock is released before reallocating it

try:
    from .config_local import *
except:
    pass



from __future__ import absolute_import
from . import core

def create_property(halo, name, prop, session):

    name = core.dictionary.get_or_create_dictionary_item(session, name)

    if isinstance(prop, core.halo.Halo):
        px = core.halo_data.HaloLink(halo, prop, name)
    else:
        px = core.halo_data.HaloProperty(halo, name, prop)

    px.creator_id = core.creator.get_creator().id
    return px


def _insert_list_unlocked(property_list):
    session = core.get_default_session()

    property_object_list = [create_property(
        p[0], p[1], p[2], session) for p in property_list if p[2] is not None]

    session.add_all(property_object_list)

    session.commit()

def insert_list(property_list):
    from tangos import parallel_tasks as pt

    if pt.backend!=None:
        with pt.RLock("insert_list"):
            _insert_list_unlocked(property_list)
    else:
        _insert_list_unlocked(property_list)





from __future__ import absolute_import
from pyramid.view import view_config
from sqlalchemy import func, and_, or_
from . import timestep_from_request

import tangos
from tangos import core

def add_urls(halos, request, sim, ts):
    for h in halos:
        h.url = request.route_url('halo_view', simid=sim.basename, timestepid=ts.extension,
                                  halonumber=h.basename)

@view_config(route_name='timestep_view', renderer='../templates/timestep_view.jinja2')
def timestep_view(request):
    ts = timestep_from_request(request)
    sim = ts.simulation

    halos = ts.halos.all()
    groups = ts.groups.all()
    trackers = ts.trackers.all()

    add_urls(halos, request, sim, ts)
    add_urls(groups, request, sim, ts)
    add_urls(trackers, request, sim, ts)

    return {'timestep': ts.extension, 'halos': halos, 'groups': groups, 'trackers': trackers,
            'gather_url': request.route_url('calculate_all',simid=request.matchdict['simid'],
                                            timestepid=request.matchdict['timestepid'],
                                            nameid="")[:-5]}



from __future__ import absolute_import
from __future__ import print_function
from sqlalchemy.orm import Session
import numpy as np
from .core import get_or_create_dictionary_item, Halo, HaloLink, TrackData
from . import query
import six
import tangos.parallel_tasks as parallel_tasks
from six.moves import zip

def generate_tracker_halo_link_if_not_present(halo_1, halo_2, dict_obj=None, weight=1.0):
    assert isinstance(halo_1, Halo)
    assert isinstance(halo_2, Halo)
    session = Session.object_session(halo_1)
    if session.query(HaloLink).filter_by(halo_from_id=halo_1.id, halo_to_id=halo_2.id).count()>0:
        return

    if dict_obj is None:
        dict_obj = get_or_create_dictionary_item(session, "tracker")
    #print halo_1.id,"->",halo_2.id
    session.merge(HaloLink(halo_1,halo_2,dict_obj,weight))

def get_trackers(sim):
    trackers = sim.trackers.all()
    nums = [tx.halo_number for tx in trackers]
    return trackers, np.array(nums)

def get_tracker_halos(ts):
    halos = ts.halos.filter_by(object_typecode=1).order_by(Halo.halo_number).all()
    hid = [h.id for h in halos]
    num = [h.halo_number for h in halos]
    return halos, np.array(num), np.array(hid)

def get_tracker_links(session, relation):
    links = session.query(HaloLink).filter_by(relation=relation).all()
    idf = [tl.halo_from_id for tl in links]
    idt = [tl.halo_to_id for tl in links]
    return links, np.array(idf), np.array(idt)

def generate_tracker_halo_links(sim, session):
    dict_obj = get_or_create_dictionary_item(session, "tracker")
    links = []
    for ts1, ts2 in parallel_tasks.distributed(list(zip(sim.timesteps[1:],sim.timesteps[:-1]))):
        print("generating links for", ts1, ts2)
        halos_1, nums1, id = get_tracker_halos(ts1)
        halos_2, nums2, id = get_tracker_halos(ts2)
        tracker_links, idf, idt = get_tracker_links(session,dict_obj)

        if len(nums1) == 0 or len(nums2) == 0:
            continue
        o1 = np.where(np.in1d(nums1,nums2))[0]
        o2 = np.where(np.in1d(nums2,nums1))[0]
        if len(o1) == 0 or len(o2) == 0:
            continue
        for ii, jj in zip(o1,o2):
            if halos_1[ii].halo_number != halos_2[jj].halo_number:
                raise RuntimeError("ERROR mismatch of BH iords")
            exists = np.where((idf==halos_1[ii].id)&(idt==halos_2[jj].id))[0]
            if len(exists) == 0:
                links.append(HaloLink(halos_1[ii],halos_2[jj],dict_obj,1.0))
                links.append(HaloLink(halos_2[jj],halos_1[ii],dict_obj,1.0))
    session.add_all(links)
    session.commit()

def new(for_simulation, using_particles):
    if isinstance(for_simulation, six.string_types):
        for_simulation = query.get_simulation(for_simulation)
    tracker = TrackData(for_simulation)
    use_iord = 'iord' in using_particles.loadable_keys()
    tracker.select(using_particles, use_iord)
    session = Session.object_session(for_simulation)
    session.add(tracker)
    tracker.create_objects()
    tracker.create_links()
    session.commit()
    return tracker.halo_number


from __future__ import absolute_import

import pypar, pypar.mpiext
import warnings
import numpy as np
from ..message import Message

def send(data, destination, tag=0):
    pypar.send(data, destination=destination, tag = tag)

def receive_any(source=None):
    if source is None:
        source = pypar.any_source
    data, status = pypar.receive(source=source, return_status=True, tag=pypar.any_tag)
    return data, status.source, status.tag

def receive(source=None, tag=0):
    if source is None:
        source = pypar.mpiext.MPI_ANY_SOURCE
    return pypar.receive(source=source,tag=tag)


def rank():
    return pypar.rank()

def size():
    return pypar.size()

def barrier():
    pypar.barrier()

def finalize():
    pypar.finalize()

class NumpyMetadataMessage(Message):
    pass

class NumpyDataMessage(Message):
    pass

def send_numpy_array(data, destination):
    pypar.send((data.shape, data.dtype),destination=destination,tag=NumpyMetadataMessage._tag)
    pypar.send(data,destination=destination,tag=NumpyDataMessage._tag,use_buffer=True,bypass=True)

def receive_numpy_array(source):
    shape,dtype = pypar.receive(source=source, tag=NumpyMetadataMessage._tag)
    ar = np.empty(shape,dtype=dtype)
    pypar.receive(source=source,buffer=ar,tag=NumpyDataMessage._tag)
    return ar



def launch(function, num_procs, args):
    if size()==1:
        raise RuntimeError("MPI run needs minimum of 2 processors (one for manager, one for worker)")
    if num_procs is not None:
        if rank()==0:
            warnings.warn("Number of processors requested (%d) will be ignored as this is an MPI run that has already selected %d processors"%(num_procs,size()))
    function(*args)


from __future__ import absolute_import
import sqlalchemy
import sqlalchemy.exc
import sqlalchemy.orm
import sqlalchemy.orm.dynamic
import sqlalchemy.orm.query
from sqlalchemy.orm import Session, contains_eager
from .. import core, temporary_halolist

class HopStrategy(object):
    """HopStrategy and its descendants define methods helpful for finding related halos, e.g. progenitors/descendants,
    or corresponding halos in other simulation runs"""

    def __init__(self, halo_from, target=None, order_by=None):
        """Construct a HopStrategy starting from the specified halo"""
        query = halo_from.links
        assert isinstance(halo_from, core.halo.Halo)
        assert isinstance(query, sqlalchemy.orm.query.Query)
        self.session = Session.object_session(halo_from)
        self.halo_from = halo_from
        self._initialise_order_by(order_by)
        self.query = query
        self._link_orm_class = core.halo_data.HaloLink
        self._target = target
        self._all = None

    def _target_timestep(self, ts):
        """Only return those hops which reach the specified timestep"""
        if ts is None:
            self.query = self.query.filter(0 == 1)
        else:
            self.query = self.query.join("halo_to").filter(core.halo.Halo.timestep_id == ts.id)

    def _target_simulation(self, sim):
        """Only return those hops which reach the specified simulation"""
        self.query = self.query.join("halo_to", "timestep").filter(
            core.timestep.TimeStep.simulation_id == sim.id)

    def _filter_query_for_target(self, db_obj):
        """Only return those hops which reach the specifid simulation or timestep"""
        if db_obj is None:
            return
        elif isinstance(db_obj, core.timestep.TimeStep):
            self._target_timestep(db_obj)
        elif isinstance(db_obj, core.simulation.Simulation):
            self._target_simulation(db_obj)
        else:
            raise ValueError("Unknown target type")

    def _initialise_order_by(self, names):
        """Specify an ordering for the output hop suggestions.

        Accepted names are:

         - 'weight' - the weight of the link, ascending (default). In the case of MultiHopStrategy, this is the
                      product of the weights along the path found.
         - 'time_asc' - the time of the snapshot, ascending order
         - 'time_desc' - the time of the snapshot, descending order
         - 'halo_number_asc' - the halo number, ascending
         - 'halo_number_desc' - the halo number, descending
         - 'nhops' - the number of hops taken to reach the halo (MultiHopStrategy only)

        Multiple names can be given to order by more than one property.
        """
        if names is None:
            names = ['weight']
        elif isinstance(names, str):
            names = [names]
        self._order_by_names = [x.lower() for x in names]

    def count(self):
        """Return the number of hops matching the conditions"""
        return len(self._get_query_all())

    def _execute_query(self):
        try:
            self._filter_query_for_target(self._target)
            results = self._query_ordered.all()
        except sqlalchemy.exc.ResourceClosedError:
            results = []

        results = [x for x in results if x is not None]

        self._all = results

    def _get_query_all(self):
        if self._all is None:
            self._execute_query()
        return self._all

    def temp_table(self):
        # N.B. this could be made more efficient
        ids_list = [x.id if hasattr(x,'id') else None for x in self.all() ]
        return temporary_halolist.temporary_halolist_table(self.session, ids_list)

    def all(self):
        """Return all possible hops matching the conditions"""
        return [x.halo_to for x in self._get_query_all()]

    def weights(self):
        """Return the weights for the possible hops"""
        return [x.weight for x in self._get_query_all()]

    def all_and_weights(self):
        """Return all possible hops matching the conditions, along with
        the weights"""
        all = self._get_query_all()
        weights = [x.weight for x in all]
        halos = [x.halo_to for x in all]
        return halos, weights

    def first(self):
        """Return the suggested hop."""
        link = self._get_query_all()
        if len(link) == 0:
            return None
        else:
            return link[0].halo_to

    def _order_by_clause(self, halo_alias, timestep_alias):
        return [self._generate_order_arg_from_name(name, halo_alias, timestep_alias) for name in self._order_by_names]

    def _generate_order_arg_from_name(self, name, halo_alias, timestep_alias):
        if name == 'weight':
            return self._link_orm_class.weight.desc()
        elif name == 'time_asc':
            return timestep_alias.time_gyr
        elif name == 'time_desc':
            return timestep_alias.time_gyr.desc()
        elif name == 'halo_number_asc':
            return halo_alias.halo_number
        elif name == 'halo_number_desc':
            return halo_alias.halo_number.desc()
        else:
            raise ValueError("Unknown ordering method %r" % name)

    def _ordering_requires_join(self):
        return 'time_asc' in self._order_by_names \
               or 'time_desc' in self._order_by_names \
               or 'halo_number_asc' in self._order_by_names \
               or 'halo_number_desc' in self._order_by_names

    @property
    def _query_ordered(self):
        query = self.query
        assert isinstance(query, sqlalchemy.orm.query.Query)
        timestep_alias = None
        halo_alias = None
        if self._ordering_requires_join():
            timestep_alias = sqlalchemy.orm.aliased(core.timestep.TimeStep)
            halo_alias = sqlalchemy.orm.aliased(core.halo.Halo)
            query = query.join(halo_alias, self._link_orm_class.halo_to).join(timestep_alias)
            query = query.options(contains_eager("halo_to", alias=halo_alias))
            query = query.options(contains_eager("halo_to", "timestep", alias=timestep_alias))
        query = query.order_by(*self._order_by_clause(halo_alias, timestep_alias))
        return query

class HopMajorDescendantStrategy(HopStrategy):
    """A hop strategy that suggests the major descendant for a halo"""

    def __init__(self, halo_from):
        target_ts = halo_from.timestep.next
        if target_ts:
            super(HopMajorDescendantStrategy, self).__init__(halo_from, target=target_ts)
        else:
            self._all = []


class HopMajorProgenitorStrategy(HopStrategy):
    """A hop strategy that suggests the major progenitor for a halo"""

    def __init__(self, halo_from):
        target_ts = halo_from.timestep.previous
        if target_ts:
            super(HopMajorProgenitorStrategy, self).__init__(halo_from, target=target_ts)
        else:
            self._all = []


#!/usr/bin/env python2.7

from __future__ import absolute_import
import sys

def run_dbwriter(argv):
    import matplotlib
    matplotlib.use('agg')
    from tangos import parallel_tasks, core
    from tangos.tools.property_writer import PropertyWriter
    writer = PropertyWriter()
    writer.parse_command_line(argv)
    parallel_tasks.launch(writer.run_calculation_loop, 2, [])

def main():
    run_dbwriter(sys.argv[1:])

if __name__ == "__main__":

    run_dbwriter(sys.argv[1:])




from __future__ import absolute_import
from . import message, log, parallel_backend_loaded
import time
import six
from ..config import DEFAULT_SLEEP_BEFORE_ALLOWING_NEXT_LOCK

class MessageRequestLock(message.Message):
    def process(self):
        lock_id = self.contents
        proc = self.source

        log.logger.debug("Received request for lock %r for proc %d", lock_id, proc)
        queue = _get_lock_queue(lock_id)
        queue.append(proc)
        if len(queue) == 1:
            _issue_next_lock(lock_id)

class MessageRelinquishLock(message.Message):
    def process(self):
        lock_id, delay_time = self.contents
        proc = self.source
        queue = _get_lock_queue(lock_id)
        assert queue[0]==proc
        queue.pop(0)
        log.logger.debug("Finished with lock %r for proc %d",lock_id, proc)
        if len(queue)>0:
            if delay_time is not None:
                time.sleep(delay_time)
            _issue_next_lock(lock_id)

class MessageGrantLock(message.Message):
    pass


_lock_queues = {}

def _get_lock_queue(lock_id):
    global _lock_queues
    lock_queue = _lock_queues.get(lock_id,[])
    _lock_queues[lock_id] = lock_queue
    return lock_queue

def _issue_next_lock(lock_id):
    queue = _get_lock_queue(lock_id)
    log.logger.debug("Issue lock %r to proc %d",lock_id, queue[0])
    MessageGrantLock(lock_id).send(queue[0])

def _any_locks_alive():
    return any([len(v)>0 for v in six.itervalues(_lock_queues)])


class RLock(object):
    def __init__(self, name, delay_before_release=DEFAULT_SLEEP_BEFORE_ALLOWING_NEXT_LOCK):
        self.name = name
        self._delay = delay_before_release
        self._count = 0

    def acquire(self):
        if not parallel_backend_loaded():
            return
        if self._count==0:
            MessageRequestLock(self.name).send(0)
            start = time.time()
            granted = MessageGrantLock.receive(0)
            assert granted.contents==self.name, "Received a lock that was not requested. The implementation of RLock is not locally thread-safe; are you using multiple threads in one process?"
            log.logger.debug("Lock %r acquired in %.1fs",self.name, time.time()-start)
        self._count+=1

    def release(self):
        if not parallel_backend_loaded():
            return
        self._count-=1
        if self._count==0:
            MessageRelinquishLock((self.name, self._delay)).send(0)

    def __enter__(self):
        self.acquire()

    def __exit__(self,type, value, traceback):
        self.release()


from __future__ import absolute_import
import numpy as np
from ..util import is_not_none


class QueryMask(object):
    """A helper class to remove rows that do not need to be queried, representing them by "None" in the final results"""
    def __init__(self):
        self.N = None

    def mark_nones_as_masked(self, input):
        """Mark any rows in the input that are None as masked, excluding them from future queries."""
        if self.N is None:
            self.N = len(input)
            to_mask = input
        else:
            to_mask = self.unmask(input)

        if len(to_mask.shape)==2:
            mask = np.any(is_not_none(to_mask),axis=0).reshape(self.N)
        elif len(to_mask.shape)==1:
            mask = is_not_none(to_mask)
        else:
            raise ValueError("Not able to use an input of this shape to determine a query mask")

        assert mask.shape==(self.N,)
        self.results_target = np.where(mask)

    def mask(self, input):
        """Mask an input array so that it only includes the rows that are to be queried"""
        self._check_ready()
        return input[self.results_target]

    def unmask(self, input, insert_value=None):
        """Unmask an output array so that it includes None for the rows that were not queried"""
        self._check_ready()
        rval_shape = list(input.shape)
        rval_shape[-1] = self.N
        rval = np.empty(rval_shape,dtype=object)
        rval[:] = insert_value
        maximal_slice = slice(None,None)
        address_tuple = tuple([maximal_slice]*(len(input.shape)-1) + [self.results_target[0]])
        rval[address_tuple] = input
        return rval

    def _check_ready(self):
        if self.N is None:
            raise RuntimeError("The query mask has not yet been configured")


from __future__ import absolute_import
from pyramid.response import Response
from pyramid.view import view_config
import socket
import tangos
from tangos import core


@view_config(route_name='simulation_list', renderer='../templates/simulation_list.jinja2')
def simulation_list(request):
    session = request.dbsession

    cats = session.query(core.DictionaryItem).join(core.simulation.SimulationProperty).all()

    titles = ["Name"] + [z.text for z in cats]
    ids = [z.id for z in cats]

    sims = session.query(tangos.core.simulation.Simulation).all()
    simulations = []
    links = []

    for x in sims:
        s = [x.basename] + ["&ndash;"] * (len(titles) - 1)

        for q in x.properties:
            s[1 + ids.index(q.name_id)] = q.data_repr()

        simulations.append(s)
        links.append(request.route_url('simulation_view',simid=x.basename))


    return {'simulations':simulations, 'titles':titles, 'links':links,
            'hostname':socket.gethostname()}




# Remote import message allows a client to ask the server to import a module.
# Currently used by the pynbody server (which cannot be loaded by default as it
# in turn imports pynbody, which cannot be a dependency)

from __future__ import absolute_import
from .message import Message
import importlib


class ImportRequestMessage(Message):
    def process(self):
        importlib.import_module(self.contents)



from .. import HaloProperties
from ...simulation_output_handlers import pynbody as pynbody_handler_module

class PynbodyHaloProperties(HaloProperties):
    works_with_handler = pynbody_handler_module.PynbodyOutputSetHandler
    requires_particle_data = True

from . import BH, zoom, centring, profile, images, gas


from __future__ import absolute_import
from . import PynbodyHaloProperties

class SphericalRegionHaloProperties(PynbodyHaloProperties):
    """A base class for calculations which require all data within a sphere (rather than just the literal
    halo finder output)"""

    def region_specification(self, db_data):
        import pynbody
        return pynbody.filt.Sphere(db_data['max_radius'], db_data['shrink_center'])

    def requires_property(self):
        return ["shrink_center", "max_radius"]+super(SphericalRegionHaloProperties, self).requires_property()




from __future__ import absolute_import
import datetime

import numpy as np
from sqlalchemy import Column, Integer, String, ForeignKey, Float, DateTime, LargeBinary
from sqlalchemy.orm import relationship, backref, Session

from . import data_attribute_mapper
from . import Base
from . import creator
from .. import simulation_output_handlers, config
from .dictionary import DictionaryItem, get_dict_id, get_or_create_dictionary_item
import six


class Simulation(Base):
    __tablename__ = 'simulations'
    # __table_args__ = {'useexisting': True}

    id = Column(Integer, primary_key=True)
    basename = Column(String)
    creator = relationship(
        creator.Creator, backref=backref('simulations', cascade='all'), cascade='save-update')
    creator_id = Column(Integer, ForeignKey('creators.id'))

    def __init__(self, basename):
        self.basename = basename
        self.creator_id = creator.get_creator().id

    def __repr__(self):
        return "<Simulation(\"" + self.basename + "\")>"

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def get_output_handler(self):
        """Get a SimulationOutputSetHandler object, pre-configured suitably for this simulation

        :rtype simulation_outputs.SimulationOutputSetHandler"""
        if not hasattr(self, "_handler"):
            handler_class = self.output_handler_class
            self._handler = handler_class(self.basename)
        return self._handler

    @property
    def output_handler_class(self):
        if not hasattr(self, "_handler_class"):
            self._handler_class = simulation_output_handlers.get_named_handler_class(self.get("handler", config.default_fileset_handler_class))
        return self._handler_class

    def keys(self):
        return [prop.name.text for prop in self.properties.all()]

    def __contains__(self, item):
        return item in list(self.keys())

    def __getitem__(self, i):
        from . import Session
        if isinstance(i, int):
            return self.timesteps[i]
        else:
            session = Session.object_session(self)
            did = get_dict_id(i, session=session)
            try:
                return session.query(SimulationProperty).filter_by(name_id=did,
                                                                   simulation_id=self.id).first().data
            except AttributeError:
                pass

        raise KeyError(i)

    def __setitem__(self, st, data):
        from . import Session
        assert isinstance(st, str)
        session = Session.object_session(self)
        name = get_or_create_dictionary_item(session, st)
        propobj = self.properties.filter_by(name_id=name.id).first()
        if propobj is None:
            propobj = session.merge(SimulationProperty(self, name, data))

        propobj.data = data
        session.commit()

    @property
    def path(self):
        return self.basename


class SimulationProperty(Base):
    __tablename__ = 'simulationproperties'

    id = Column(Integer, primary_key=True)
    name_id = Column(Integer, ForeignKey('dictionary.id'))
    name = relationship(DictionaryItem)

    simulation_id = Column(Integer, ForeignKey('simulations.id'))
    simulation = relationship(Simulation, backref=backref('properties', cascade='all, delete-orphan',
                                                          lazy='dynamic', order_by=name_id), cascade='save-update')

    creator_id = Column(Integer, ForeignKey('creators.id'))
    creator = relationship(creator.Creator, backref=backref(
        'simproperties', cascade='all, delete', lazy='dynamic'), cascade='save-update')

    data_float = Column(Float)
    data_int = Column(Integer)
    data_time = Column(DateTime)
    data_string = Column(String)
    data_array = Column(LargeBinary)


    def __init__(self, sim, name, data):
        self.simulation = sim
        if not isinstance(name, DictionaryItem):
            name = get_or_create_dictionary_item(Session.object_session(self), name)
        self.name = name
        self.data = data
        self.creator_id = creator.get_creator().id

    def data_repr(self):
        f = self.data
        if type(f) is float:
            x = "%.2g" % f
        elif type(f) is datetime.datetime:
            x = f.strftime('%H:%M %d/%m/%y')
        elif type(f) is str or type(f) is six.text_type:
            x = "'%s'" % f
        elif f is None:
            x = "None"
        elif isinstance(f, np.ndarray):
            x = str(f)
        else:
            x = "%d" % f

        return x

    def __repr__(self):
        x = "<SimulationProperty " + self.name.text + \
            " of " + self.simulation.__repr__()
        x += " = " + self.data_repr()
        x += ">"
        return x

    @property
    def data(self):
        return data_attribute_mapper.get_data_of_unknown_type(self)

    @data.setter
    def data(self, data):
        data_attribute_mapper.set_data_of_unknown_type(self, data)


from __future__ import absolute_import
import contextlib, weakref, sys, gc
from ..log import logger

@contextlib.contextmanager
def check_deleted(a):
    if a is None:
        yield
        return
    else:
        a_s = weakref.ref(a)
        #sys.exc_clear()
        del a
        yield
        gc.collect()
        if a_s() is not None:
            logger.error("check_deleted failed")
            logger.error("gc reports hanging references: %s", gc.get_referrers(a_s()))


from __future__ import absolute_import
from . import LiveHaloProperties

class IntrinsicProperties(LiveHaloProperties):
    names = "t","z","a","dbid", "halo_number", "NDM", "NStar", "NGas", "type", "step_path"

    def live_calculate(self, halo):
        ts = halo.timestep
        return ts.time_gyr, ts.redshift, 1./(1.+ts.redshift), halo.id, halo.halo_number, \
               halo.NDM, halo.NStar, halo.NGas, halo.object_typecode, str(halo.timestep.path)


#!/usr/bin/env python2.7


from __future__ import absolute_import
from tangos import core
from tangos import parallel_tasks
from tangos.parallel_tasks import database
from tangos.tools.crosslink import CrossLinker
import sys


def run_dbwriter(argv):
    database.synchronize_creator_object()
    writer = CrossLinker()
    writer.parse_command_line(argv)
    writer.run_calculation_loop()


def main():
    parallel_tasks.launch(run_dbwriter, 2, [sys.argv[1:]])




from __future__ import absolute_import
from sqlalchemy.orm import relationship



from .property import HaloProperty
from .link import HaloLink



def _initialise_halo_property_relationships():

    from ..halo import Halo
    from ..timestep import TimeStep

    Halo.properties = relationship(HaloProperty, cascade='all', lazy='dynamic',
                                   primaryjoin=(HaloProperty.halo_id == Halo.id) & (
                                       HaloProperty.deprecated == False),
                                   uselist=True)


    Halo.deprecated_properties = relationship(HaloProperty, cascade='all',
                                              primaryjoin=(HaloProperty.halo_id == Halo.id) & (
                                                  HaloProperty.deprecated == True),
                                              uselist=True)

    # eager loading support:

    #Halo.all_properties = relationship(HaloProperty, primaryjoin=(HaloProperty.halo_id == Halo.id) & (
    #                                  HaloProperty.deprecated == False))


    TimeStep.links_from = relationship(HaloLink, secondary=Halo.__table__,
                                       secondaryjoin=(
                                           HaloLink.halo_from_id == Halo.id),
                                       primaryjoin=(
                                           Halo.timestep_id == TimeStep.id),
                                       cascade='none', lazy='dynamic')



    TimeStep.links_to = relationship(HaloLink, secondary=Halo.__table__,
                                     secondaryjoin=(
                                         HaloLink.halo_to_id == Halo.id),
                                     primaryjoin=(Halo.timestep_id == TimeStep.id),
                                     cascade='none', lazy='dynamic')



    Halo.all_links = relationship(HaloLink, primaryjoin=(HaloLink.halo_from_id == Halo.id))
    Halo.all_reverse_links = relationship(HaloLink, primaryjoin=(HaloLink.halo_to_id == Halo.id))

_initialise_halo_property_relationships()


from __future__ import division
from ..config import mergertree_timeout, mergertree_max_nhalos, mergertree_min_fractional_NDM, mergertree_min_fractional_weight
from . import MultiSourceMultiHopStrategy
from ..util import consistent_collection as cc
from .. import core
import time, math
from six.moves import range
from six.moves import zip


class MergerTree(object):
    """Construct a merger tree from a given starting halo.

    This is chiefly for use by the web interface, since displaying the resulting data
    is non-trivial. However, it is also possible to use MergerTree from an interactive
    python session as follows:

     halo = tangos.get_halo(...)
     tree = tangos.relation_finding.tree.MergerTree(halo)
     tree.construct()
     tree.plot()

    This will display the tree information in matplotlib."""

    def __init__(self, base_halo):
        """Initialise the tree starting at the specified base halo.

        Note that the method construct() must be called to actually build the tree"""
        self.base_halo = base_halo
        self.highlight_halo = base_halo
        self.timeout = mergertree_timeout
        self.must_include = []
        self.x_step = 5

    def construct(self):
        """Construct the tree"""
        self._construction_start_time = time.time()
        self._treedata = self._construct_preliminary([self.base_halo])[0]
        self._postprocess()

    def _construct_preliminary(self, halos, depth=0):
        """Construct a preliminary representation of the tree, which will later be revised
        by the post-processing"""
        if len(halos) == 0:
            return []

        visited = []

        output = [self._get_basic_halo_node(halo, depth) for halo in halos]

        previous_timestep = cc.ConsistentCollection(halos).timestep.previous
        if previous_timestep is not None:

            rl = MultiSourceMultiHopStrategy(halos,target=previous_timestep, nhops_max=1,
                                             one_match_per_input=False)

            link_objs = rl._get_query_all()
        else:
            link_objs = []

        pairings = []
        next_level_halos = []
        next_level_halos_link_from = []

        maxdepth = 0

        if len(link_objs) > 0 and time.time() - self._construction_start_time < self.timeout:
            max_weight = max([o.weight for o in link_objs])
            NDM_array = [o.halo_to.NDM for o in link_objs]
            max_NDM = max(NDM_array)
            if len(NDM_array) > mergertree_max_nhalos:
                NDM_cut = sorted(NDM_array)[mergertree_max_nhalos]
            else:
                NDM_cut = None

            for obj in link_objs:
                should_construct_onward_tree = obj.weight > max_weight * mergertree_min_fractional_weight
                should_construct_onward_tree &= obj.halo_to.NDM > mergertree_min_fractional_NDM * max_NDM
                if NDM_cut:
                    should_construct_onward_tree &= obj.halo_to.NDM > NDM_cut

                if obj.halo_to_id in self.must_include:
                    should_construct_onward_tree = True  # override normal criteria
                if obj.halo_to_id in visited:
                    should_construct_onward_tree = False  # already expanded via another link

                if should_construct_onward_tree:
                    next_level_halos.append(obj.halo_to)
                    next_level_halos_link_from.append(halos.index(obj.halo_from))
                    visited.append(obj.halo_to_id)

            next_level_items = self._construct_preliminary(next_level_halos,  depth + 1)

            for placement_id, item in zip(next_level_halos_link_from, next_level_items):
                output[placement_id]['contents'].append(item)
                if item['maxdepth'] > maxdepth:
                    maxdepth = item['maxdepth']

        for i in output:
            i['maxdepth'] = maxdepth + 1

        return output


    def _get_basic_halo_node(self, halo, depth):
        """Get the dictionary of properties belonging to a node of the tree, i.e. a given halo"""
        timeinfo = "TS ...%s; z=%.2f; t=%.2e Gyr" % (
            halo.timestep.extension[-5:], halo.timestep.redshift, halo.timestep.time_gyr)

        if halo.NDM > 0:
            moreinfo = "%s %d, NDM=%.2e" % (halo.__class__.__name__, halo.halo_number, halo.NDM)
            unscaled_size = math.log10(halo.NDM)
        else:
            moreinfo = "%s %d" % (halo.__class__.__name__, halo.halo_number)
            unscaled_size = 1

        try:
            Mvir = halo.properties.filter_by(
                name_id=core.dictionary.get_dict_id("Mvir", session=core.Session.object_session(halo),
                                                           allow_query=False)).first()
        except KeyError:
            Mvir = None
        if Mvir is not None:
            moreinfo += ", %s=%.2e" % ("Mvir", Mvir.data)

        nodeclass = 'node-dot-standard'
        name = str(halo.halo_number)

        if halo == self.highlight_halo:
            nodeclass = 'node-dot-selected'
        elif depth == 0:
            if halo.next is not None:
                nodeclass = 'node-dot-continuation'
                name = '...'
                moreinfo = "Continues... " + moreinfo
        if len(name) > 4:
            name = ""
        output = {'name': name,
                  'nodeclass': nodeclass,
                  'moreinfo': moreinfo,
                  'timeinfo': timeinfo,
                  'halo_number': halo.halo_number,
                  'unscaled_size': unscaled_size,
                  'contents': [],
                  'depth': depth}
        return output

    def plot(self):
        """Display the tree in matplotlib"""
        import pylab as p
        ax = p.gca()
        for node in self._visit_tree():
            y = node['depth']*self.x_step
            x = node['_x']
            circ = p.Circle((x,y),node['size'],facecolor='black',edgecolor='none')
            ax.add_patch(circ)
            ax.text(x+node['size']*1.1,y,node['name'])
            for child_node in node['contents']:
                p.plot([node['_x'],child_node['_x']],[y,y+self.x_step],'k')
        ax.set_aspect('equal','datalim')


    def __str__(self):
        result = ""
        for layer in self._visit_tree_layers():
            max_x = max([node['_x'] for node in layer])
            this_layer_stubs = this_layer_connections = this_layer_string = " "*(max_x+self.x_step)
            for node in layer:
                x0 = node['_x']
                x1 = x0+self.x_step
                node_string = str(node['halo_number']).center(self.x_step)
                this_layer_string=this_layer_string[:x0]+node_string+this_layer_string[x1:]

                if 'contents' in node and len(node['contents'])>0 :
                    this_layer_stubs = this_layer_stubs[:x0+self.x_step//2]+"|"+this_layer_stubs[x0+self.x_step//2:]
                    x0_min_next_layer = min([child_node['_x'] for child_node in node['contents']])
                    x0_max_next_layer = max([child_node['_x'] for child_node in node['contents']])
                    line_start = x0_min_next_layer+self.x_step//2
                    line_finish = x0_max_next_layer+self.x_step//2
                    line = "-"*(line_finish-line_start)
                    this_layer_connections = this_layer_connections[:line_start] + line +\
                                             this_layer_connections[line_finish:]

            result+=this_layer_string+"\r\n"+this_layer_stubs+"\r\n"+this_layer_connections+"\r\n"
        return result


    def _visit_tree(self, tree=None):
        """Yields each tree element in turn"""
        if tree is None:
            tree = self._treedata
        yield tree
        for subtree in tree['contents']:
            for item in self._visit_tree(subtree):
                yield item

    def _visit_tree_layers(self):
        """Yields each layer of the tree in turn, consisting of a list of elements at each layer"""
        layers = {}
        for leaf in self._visit_tree():
            this_depth = leaf['depth']
            if this_depth not in layers:
                layers[this_depth] = []
            layers[this_depth].append(leaf)
        for i in range(self._treedata['maxdepth']):
            yield layers[i]


    def _postprocess_megertree_rescale(self):
        max_size = -100
        for node in self._visit_tree():
            if node['unscaled_size'] > max_size:
                max_size = node['unscaled_size']

        for node in self._visit_tree():
            size = self.x_step/3 + (self.x_step/10) * (node['unscaled_size'] - max_size)
            if size < self.x_step/10:
                size = self.x_step/10
            node['size'] = size

    def _postprocess(self):
        """Once all tree information is present, re-process the nodes to a more sensible layout"""
        self._postprocess_megertree_rescale()
        self._postprocess_mergertree_layout_by_branch()

    def _postprocess_mergertree_layout_by_branch(self):
        """Associate with each tree node a toplogical spatial range such that branches do not cross"""
        self._treedata['space_range'] = (0.0, 1.0)
        existing_ranges = [{} for i in range(self._treedata['maxdepth'])]
        for node in self._visit_tree():
            x_start, x_end = node['space_range']
            node['mid_range'] = (x_start + x_end) / 2
            if len(node['contents']) > 0:
                delta = (x_end - x_start) / len(node['contents'])
                total_nodes = len(node['contents'])
                halo_numbers = [child['halo_number'] for child in node['contents']]
                halo_numbers.sort()
                for i, child in enumerate(node['contents']):
                    # create an index that starts in the middle then works outwards
                    rank = halo_numbers.index(child['halo_number'])
                    sign = 2 * ((rank + 1) % 2) - 1
                    i_shuffled = total_nodes // 2 + sign * ((rank + 1) // 2)
                    child_range = existing_ranges[child['depth']].get(child['halo_number'],
                                                                      (x_start + i_shuffled * delta,
                                                                       x_start + (i_shuffled + 1) * delta))

                    child['space_range'] = child_range
                    existing_ranges[child['depth']][child['halo_number']] = child_range

        self._postprocess_mergertree_layout_by_number('mid_range')

    def _postprocess_mergertree_layout_by_number(self, key='halo_number'):
        """Associate a position value _x to each node that increases with the specified key

        The _x value is guaranteed to increase with increasing key value but also will be centred on zero"""
        x_vals = [set() for i in range(self._treedata['maxdepth'])]

        for node in self._visit_tree():
            x_vals[node['depth']].add(node[key])

        max_entries = max([len(v) for v in x_vals])
        x_map = [{} for i in range(self._treedata['maxdepth'])]
        for this_vals, this_map in zip(x_vals, x_map):
            new_x = (self.x_step//2) * (max_entries - len(this_vals))
            for xv in sorted(this_vals):
                this_map[xv] = new_x
                new_x += self.x_step

        for node in self._visit_tree():
            node['_x'] = x_map[node['depth']][node[key]]
